# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import logging
import os
from gettext import gettext as _
from gettext import pgettext as C_
from pathlib import Path

import tmdbsimple
from gi.repository import Adw, Gio, GLib, GObject, Gtk
from requests import HTTPError

from . import shared  # type: ignore
from .background_queue import ActivityType, BackgroundActivity, BackgroundQueue
from .models.language_model import LanguageModel
from .providers.local_provider import LocalProvider as local
from .providers.tmdb_provider import TMDBProvider as tmdb


@Gtk.Template(resource_path=shared.PREFIX + '/ui/preferences.ui')
class PreferencesDialog(Adw.PreferencesDialog):
    __gtype_name__ = 'PreferencesDialog'

    _download_group = Gtk.Template.Child()
    _language_comborow = Gtk.Template.Child()
    _language_model = Gtk.Template.Child()
    _update_freq_comborow = Gtk.Template.Child()
    _offline_group = Gtk.Template.Child()
    _offline_switch = Gtk.Template.Child()
    _tmdb_group = Gtk.Template.Child()
    _use_own_key_switch = Gtk.Template.Child()
    _own_key_entryrow = Gtk.Template.Child()
    _housekeeping_group = Gtk.Template.Child()
    _exit_cache_switch = Gtk.Template.Child()
    _cache_row = Gtk.Template.Child()
    _data_row = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        self.language_change_handler = self._language_comborow.connect(
            'notify::selected', self._on_language_changed)
        self._update_freq_comborow.connect(
            'notify::selected', self._on_freq_changed)

        shared.schema.bind('onboard-complete', self._offline_group,
                           'sensitive', Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('onboard-complete', self._download_group,
                           'visible', Gio.SettingsBindFlags.INVERT_BOOLEAN)
        shared.schema.bind('onboard-complete', self._tmdb_group,
                           'sensitive', Gio.SettingsBindFlags.DEFAULT)

        shared.schema.bind('offline-mode', self._offline_switch,
                           'active', Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('use-own-tmdb-key', self._use_own_key_switch,
                           'active', Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('exit-remove-cache', self._exit_cache_switch,
                           'active', Gio.SettingsBindFlags.DEFAULT)

        self._offline_switch.connect('notify::active', lambda pspec, user_data: logging.debug(
            f'Toggled offline mode: {self._offline_switch.get_active()}'))
        self._use_own_key_switch.connect(
            'notify::active', self._on_use_own_key_switch_activated)
        self._exit_cache_switch.connect('notify::active', lambda pspec, user_data: logging.debug(
            f'Toggled clear cache on exit: {self._exit_cache_switch.get_active()}'))

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for the "map" signal.
        Populates dropdowns and checks if an automatic update of the content is due.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        if shared.schema.get_boolean('onboard-complete'):
            self._setup_languages()

        # Update frequency dropdown
        match shared.schema.get_string('update-freq'):
            case 'never':
                self._update_freq_comborow.set_selected(0)
            case 'day':
                self._update_freq_comborow.set_selected(1)
            case 'week':
                self._update_freq_comborow.set_selected(2)
            case 'month':
                self._update_freq_comborow.set_selected(3)

        self._own_key_entryrow.set_text(
            shared.schema.get_string('own-tmdb-key'))

        # Update occupied space
        self._update_occupied_space()

    def _setup_languages(self):
        self._language_comborow.handler_block(self.language_change_handler)

        languages = local.get_all_languages()
        languages.pop(len(languages)-6)    # remove 'no language'
        for language in languages:
            self._language_model.append(language.name)

        self._language_comborow.set_selected(
            self._get_selected_language_index(shared.schema.get_string('tmdb-lang')))
        self._language_comborow.handler_unblock(self.language_change_handler)

    def _on_language_changed(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for "notify::selected" signal.
        Updates the prefered TMDB language in GSettings.

        Args:
            pspec (GObject.ParamSpec): The GParamSpec of the property which changed
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        language = self._get_selected_language(
            self._language_comborow.get_selected_item().get_string())
        shared.schema.set_string('tmdb-lang', language)
        logging.debug(f'Changed TMDB language to {language}')

    def _on_freq_changed(self, pspec, user_data: object | None) -> None:
        """
        Callback for "notify::selected" signal.
        Updates the frequency for content updates in GSettings.

        Args:
            pspec (GObject.ParamSpec): The GParamSpec of the property which changed
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        freq = self._update_freq_comborow.get_selected()
        match freq:
            case 0:
                shared.schema.set_string('update-freq', 'never')
                logging.debug(f'Changed update frequency to never')
            case 1:
                shared.schema.set_string('update-freq', 'day')
                logging.debug(f'Changed update frequency to day')
            case 2:
                shared.schema.set_string('update-freq', 'week')
                logging.debug(f'Changed update frequency to week')
            case 3:
                shared.schema.set_string('update-freq', 'month')
                logging.debug(f'Changed update frequency to month')

    def _get_selected_language_index(self, iso_name: str) -> int:
        """
        Loops all available languages and returns the index of the one with the specified iso name. If a result is not found, it returns the index for English (37).

        Args:
            iso_name: a language's iso name

        Return:
            int with the index
        """

        for idx, language in enumerate(local.get_all_languages()):
            if language.iso_name == iso_name:
                return idx
        return 37

    def _get_selected_language(self, name: str) -> str:
        """
        Loops all available languages and returns the iso name of the one with the specified name. If a result is not found, it returns the iso name for English (en).

        Args:
            name: a language's name

        Return:
            str with the iso name
        """

        for language in local.get_all_languages():
            if language.name == name:
                return language.iso_name
        return 'en'

    @Gtk.Template.Callback('_on_download_activate')
    def _on_download_activate(self, user_data: object | None) -> None:
        """
        Completes the downlaod, stores the data in the db and sets the relevant GSettings.

        Args:
            None

        Results:
            None
        """

        logging.info('Attempting first setup completetion')
        Gio.NetworkMonitor.get_default().can_reach_async(
            Gio.NetworkAddress.parse_uri('https://api.themoviedb.org', 80),
            None,
            self._on_reach_done,
            None
        )

    def _on_reach_done(self, source: GObject.Object | None, result: Gio.AsyncResult, data: object | None) -> None:
        """
        Callback for asynchronous network check, reached after the first call.
        If the network is available, proced with the download, otherwise keep checking every second until it becomes
        available or the user goes in offline mode.

        Args:
            source (GObject.Object or None): the object the asynchronous operation was started with.
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        try:
            network = Gio.NetworkMonitor.get_default().can_reach_finish(result)
        except GLib.Error:
            network = None

        if network:
            logging.error('Network ok, start download')
            languages = tmdb.get_languages()
            for lang in languages:
                local.add_language(LanguageModel(lang))
                
            local.update_movies_table()
            local.update_series_table()

            shared.schema.set_boolean('first-run', False)
            shared.schema.set_boolean('offline-mode', False)
            shared.schema.set_boolean('onboard-complete', True)
            shared.schema.set_boolean('db-needs-update', False)

            self._setup_languages()
            
            Gio.NetworkMonitor.get_default().connect(
                'network-changed', self._on_network_changed)
        else:
            logging.error('No network, aborting first setup completion')
            dialog = Adw.AlertDialog.new(
                heading=C_('message dialog heading', 'No Network'),
                body=C_('message dialog body', 'Connect to the Internet to complete the setup.')
            )
            dialog.add_response('ok', C_('alert dialog action', '_OK'))
            dialog.choose(self, None, None, None)

    def _on_network_changed(self, network_monitor: Gio.NetworkMonitor, network_available: bool) -> None:
        """
        Callback for "network-changed" signal.
        If no network is available, it turns on offline mode.

        Args:
            network_monitor (Gio.NetworkMonitor): the NetworkMonitor in use
            network_available (bool): whether or not the network is available

        Returns:
            None
        """

        shared.schema.set_boolean(
            'offline-mode', GLib.Variant.new_boolean(not network_available))

    @Gtk.Template.Callback('_on_clear_cache_activate')
    def _on_clear_cache_activate(self, user_data: object | None) -> None:
        """
        Callback for "activated" signal.
        Shows a confirmation dialog to the user.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        logging.debug('Show cache clear dialog')
        builder = Gtk.Builder.new_from_resource(
            shared.PREFIX + '/ui/dialogs/message_dialogs.ui')
        _clear_cache_dialog = builder.get_object('_clear_cache_dialog')
        _clear_cache_dialog.choose(self,
            None, self._on_cache_message_dialog_choose, None)

    def _on_cache_message_dialog_choose(self,
                                        source: GObject.Object | None,
                                        result: Gio.AsyncResult,
                                        user_data: object | None) -> None:
        """
        Callback for the message dialog.
        Finishes the async operation and retrieves the user response. If the later is positive, adds a background activity to delete the cache.

        Args:
            source (Gtk.Widget): object that started the async operation
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        result = Adw.AlertDialog.choose_finish(source, result)
        if result == 'cache_cancel':
            logging.debug('Cache clear dialog: cancel, aborting')
            return

        logging.debug('Cache clear dialog: sel, aborting')
        BackgroundQueue.add(
            activity=BackgroundActivity(
                activity_type=ActivityType.REMOVE,
                title=C_('Background activity title', 'Clear cache'),
                task_function=self._clear_cache),
            on_done=self._on_cache_clear_done)

    def _clear_cache(self, activity: BackgroundActivity) -> None:
        """
        Clears the cache.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """

        logging.info('Deleting cache')
        files = glob.glob('*.jpg', root_dir=shared.cache_dir)
        for file in files:
            os.remove(shared.cache_dir / file)
            logging.debug(f'Deleted {shared.cache_dir / file}')

    def _on_cache_clear_done(self,
                             source: GObject.Object,
                             result: Gio.AsyncResult,
                             cancellable: Gio.Cancellable,
                             activity: BackgroundActivity):
        self._update_occupied_space()
        activity.end()

    @Gtk.Template.Callback('_on_clear_activate')
    def _on_clear_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "activated" signal.
        Shows a confirmation dialog to the user.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        logging.debug('Show data clear dialog')
        builder = Gtk.Builder.new_from_resource(
            shared.PREFIX + '/ui/dialogs/message_dialogs.ui')
        _clear_data_dialog = builder.get_object('_clear_data_dialog')
        _movies_row = builder.get_object('_movies_row')
        _series_row = builder.get_object('_series_row')
        self._movies_checkbtn = builder.get_object('_movies_checkbtn')
        self._series_checkbtn = builder.get_object('_series_checkbtn')

        # TRANSLATORS: {number} is the number of titles
        _movies_row.set_subtitle(_('{number} Titles').format(
            number=len(local.get_all_movies())))
        _series_row.set_subtitle(_('{number} Titles').format(
            number=len(local.get_all_series())))

        _clear_data_dialog.choose(self,
            None, self._on_data_message_dialog_choose, None)

    def _on_data_message_dialog_choose(self,
                                       source: GObject.Object | None,
                                       result: Gio.AsyncResult,
                                       user_data: object | None) -> None:
        """
        Callback for the message dialog.
        Finishes the async operation and retrieves the user response. If the later is positive, adds background activities to delete the selected data.

        Args:
            source (Gtk.Widget): object that started the async operation
            result (Gio.AsyncResult): a Gio.AsyncResultresult = Adw.AlertDialog.choose_finish(source, result)
        if result == 'data_cancel':
            return

            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        result = Adw.AlertDialog.choose_finish(source, result)
        if result == 'data_cancel':
            logging.debug('Data clear dialog: cancel, aborting')
            return

        # Movies
        if self._movies_checkbtn.get_active():
            logging.debug('Data clear dialog: movies selected')
            BackgroundQueue.add(
                activity=BackgroundActivity(
                    activity_type=ActivityType.REMOVE,
                    title=C_('Background activity title', 'Delete all movies'),
                    task_function=self._clear_movies),
                on_done=self._on_data_clear_done)

        # TV Series
        if self._series_checkbtn.get_active():
            logging.debug('Data clear dialog: tv series selected')
            BackgroundQueue.add(
                activity=BackgroundActivity(
                    activity_type=ActivityType.REMOVE,
                    title=C_('Background activity title',
                             'Delete all TV Series'),
                    task_function=self._clear_series),
                on_done=self._on_data_clear_done)

    def _clear_movies(self, activity: BackgroundActivity) -> None:
        """
        Clears all movies.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """

        logging.info('Deleting all movies')
        for movie in local.get_all_movies():    # type: ignore
            local.delete_movie(movie.id)
            logging.debug(f'Deleted ({movie.id}) {movie.title}')

    def _on_data_clear_done(self,
                            source: GObject.Object,
                            result: Gio.AsyncResult,
                            cancellable: Gio.Cancellable,
                            activity: BackgroundActivity):
        """Callback to complete async activity"""

        self._update_occupied_space()
        self.get_parent().activate_action('win.refresh', None)
        activity.end()

    def _clear_series(self, activity: BackgroundActivity) -> None:
        """
        Clears all TV series.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """

        logging.info('Deleting all TV series')
        for serie in local.get_all_series():    # type: ignore
            local.delete_series(serie.id)
            logging.debug(f'Deleted ({serie.id}) {serie.title}')

    def _calculate_space(self, directory: Path) -> float:
        """
        Given a directory, calculates the total space occupied on disk.

        Args:
            directory (Path): the directory to measure

        Returns:
            float with space occupied in MegaBytes (MB)
        """

        return sum(file.stat().st_size for file in directory.rglob('*'))/1024.0/1024.0

    def _update_occupied_space(self) -> None:
        """
        After calculating space occupied by cache and data, updates the ui labels to reflect the values.

        Args:
            None

        Returns:
            None
        """

        cache_space = self._calculate_space(shared.cache_dir)
        data_space = self._calculate_space(shared.data_dir)

        self._housekeeping_group.set_description(  # TRANSLATORS: {total_space:.2f} is the total occupied space
            _('Ticket Booth is currently using {total_space:.2f} MB. Use the options below to free some space.').format(total_space=cache_space+data_space))

        # TRANSLATORS: {space:.2f} is the occupied space
        self._cache_row.set_subtitle(
            _('{space:.2f} MB occupied').format(space=cache_space))
        self._data_row.set_subtitle(
            _('{space:.2f} MB occupied').format(space=data_space))

    @Gtk.Template.Callback('_on_own_key_changed')
    def _on_own_key_changed(self, user_data: object | None) -> None:
        """
        Callback for 'changed' signal.
        Removes all CSS classes.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._own_key_entryrow.remove_css_class('error')
        self._own_key_entryrow.remove_css_class('success')

    @Gtk.Template.Callback('_on_check_own_key_button_clicked')
    def _on_check_own_key_button_clicked(self, user_data: object | None) -> None:
        """
        Callback for 'clicked' signal.
        Temporarelly changes the API key in use with the one provided and checks if it is valid.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        tmdb.set_key(self._own_key_entryrow.get_text())
        try:
            tmdb.get_movie(575264)
            shared.schema.set_string(
                'own-tmdb-key', self._own_key_entryrow.get_text())
            self._own_key_entryrow.add_css_class('success')
        except (tmdbsimple.base.APIKeyError, HTTPError):
            self._own_key_entryrow.add_css_class('error')
            tmdb.set_key(tmdb.get_builtin_key())

    def _on_use_own_key_switch_activated(self, pspec, user_data: object | None) -> None:
        """
        Callback for 'notify::active' signal.
        Sets the API key is use based on the state of the property.

        Args:
            pspec (GObject.ParamSpec): pspec of the changed property
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        if self._use_own_key_switch.get_active():
            tmdb.set_key(self._own_key_entryrow.get_text())
        else:
            tmdb.set_key(tmdb.get_builtin_key())



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import logging
import os
from gettext import gettext as _
from gettext import pgettext as C_

from gi.repository import Adw, Gio, GLib, Gtk

from . import shared  # type: ignore
from .background_queue import BackgroundQueue
from .dialogs.add_manual_dialog import AddManualDialog
from .dialogs.add_tmdb_dialog import AddTMDBDialog
from .views.first_run_view import FirstRunView
from .views.db_update_view import DbUpdateView
from .views.main_view import MainView
from .providers.local_provider import LocalProvider as local


@Gtk.Template(resource_path=shared.PREFIX + '/ui/window.ui')
class TicketboothWindow(Adw.ApplicationWindow):
    """
    This class reppresents the main application window.

    Properties:
        None

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'TicketboothWindow'

    _win_stack = Gtk.Template.Child()

    def _sort_on_changed(self, new_state: str, source: Gtk.Widget) -> None:
        """
        Callback for the win.view-sorting action

        Args:
            new_state (str): new selected state
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        self.set_state(new_state)
        logging.debug(f'Sort: {new_state}')
        shared.schema.set_string('view-sorting', str(new_state)[1:-1])

    def _add_tmdb(self, new_state: None, source: Gtk.Widget) -> None:
        """
        Callback for the win.add-tmdb action

        Args:
            new_state (None): stateless action, always None

        Returns:
            None
        """

        dialog = AddTMDBDialog()
        logging.info('Add from TMDB dialog open')
        dialog.present(source)

    def _add_manual(self, new_state: None, source: Gtk.Widget) -> None:
        """
        Callback for the win.add-manual action

        Args:
            new_state (None): stateless action, always None

        Returns:
            None
        """

        dialog = AddManualDialog()
        logging.info('Add manual dialog open')
        dialog.present(source)

    def _refresh(self, new_state: None, source: Gtk.Widget) -> None:
        """
        Callback for the win.refresh action

        Args:
            new_state (None): stateless action, always None
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        logging.info('Refresh requested')
        source._win_stack.get_child_by_name('main').refresh()

    def _update_background_indicator(self, new_state: None, source: Gtk.Widget) -> None:
        """
        Callback for the win.update-background-indicator

        Args:
            new_state (None): stateless action, always None
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        source._win_stack.get_child_by_name(
            'main')._background_indicator.refresh()

    def _unwatched_first_changed(self, new_state: GLib.Variant, source: Gtk.Widget) -> None:
        """
        Callback for the win.unwatched-first action

        Args:
            new_state (bool): new selected state
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        logging.debug(f'Sort unwatched first: {new_state.get_boolean()}')
        shared.schema.set_boolean('unwatched-first', new_state.get_boolean())
        self.set_state(new_state)

    def _separate_watched_changed(self, new_state: GLib.Variant, source: Gtk.Widget) -> None:
        """
        Callback for the win.separate-watched action

        Args:
            new_state (bool): new selected state
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        logging.debug(f'Separate watched: {new_state.get_boolean()}')
        shared.schema.set_boolean('separate-watched', new_state.get_boolean())
        self.set_state(new_state)

    def _hide_watched_changed(self, new_state: GLib.Variant, source: Gtk.Widget) -> None:
        """
        Callback for the win.hide-watched action

        Args:
            new_state (bool): new selected state
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        logging.debug(f'Hide watched: {new_state.get_boolean()}')
        shared.schema.set_boolean('hide-watched', new_state.get_boolean())
        self.set_state(new_state)
        
    def _search(self, new_state: GLib.Variant, source: Gtk.Widget) -> None:
        """
        Callback for the win.search action

        Args:
            new_state (bool): new selected state
            source (Gtk.Widget): widget that caused the activation

        Returns:
            None
        """

        logging.debug(f'Search: {new_state.get_boolean()}')
        shared.schema.set_boolean('search-enabled', new_state.get_boolean())
        self.set_state(new_state)

    _actions = {
        ('view-sorting', None, 's',
         f"'{shared.schema.get_string('view-sorting')}'", _sort_on_changed),
        ('separate-watched', None, None, 'true' if shared.schema.get_boolean(
            'separate-watched') else 'false', _separate_watched_changed),
        ('hide-watched', None, None, 'true' if shared.schema.get_boolean('hide-watched')
         else 'false', _hide_watched_changed),
        ('add-tmdb', _add_tmdb),
        ('add-manual', _add_manual),
        ('refresh', _refresh),
        ('update-backgroud-indicator', _update_background_indicator),
        ('search', None, None, 'true' if shared.schema.get_boolean('search-enabled')
         else 'false', _search),
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_action_entries(self._actions, self)
        self._restore_state()
        self.app = kwargs.get("application")
        if shared.DEBUG:
            self.add_css_class('devel')

        shared.schema.bind('offline-mode', self.lookup_action('add-tmdb'),
                           'enabled', Gio.SettingsBindFlags.INVERT_BOOLEAN)
        shared.schema.bind('separate-watched', self.lookup_action('hide-watched'),
                           'enabled', Gio.SettingsBindFlags.INVERT_BOOLEAN)

        if shared.schema.get_boolean('onboard-complete'):
            Gio.NetworkMonitor.get_default().connect(
                'network-changed', self._on_network_changed)

    @Gtk.Template.Callback('_on_close_request')
    def _on_close_request(self, user_data: object | None) -> bool:
        """
        Callback for "close-request" signal.
        Checks for background activities to prevent quiting and corruption, deletes cached data if enabled in settings.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            True to block quiting, False to allow it
        """

        logging.info('Close requested')

        # Background activities
        if not all(activity.completed for activity in BackgroundQueue.get_queue()):
            dialog = Adw.AlertDialog.new(
                heading=C_('message dialog heading', 'Background Activies Running'),
                body=C_('message dialog body', 'Some activities are running in the background and need to be completed before exiting. Look for the indicator in the header bar to check when they are finished.')
            )
            dialog.add_response('ok', C_('alert dialog action', '_OK'))
            dialog.choose(self, None, None, None)
            logging.error('Close inhibited, running activities in background')
            return True

        # Cache
        if shared.schema.get_boolean('exit-remove-cache'):
            files = glob.glob('*.jpg', root_dir=shared.cache_dir)
            for file in files:
                os.remove(shared.cache_dir / file)
            logging.info('Cache deleted')

        # recent_change reset
        local.reset_recent_change()
        logging.info('recent_change reseted')
        logging.info('Closing')
        return False

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, widget: Gtk.Widget) -> None:
        """
        Callback for the "map" signal. Determines what view to show on startup.

        Args:
            widget (Gtk.Widget): the object which received the signal

        Returns:
            None
        """

        is_first_run = shared.schema.get_boolean('first-run')
        logging.info(f'is first run: {is_first_run}')
        if is_first_run:
            logging.info('Start first run setup')
            self.first_run_view = FirstRunView()
            self._win_stack.add_named(child=self.first_run_view, name='first-run')
            self._win_stack.set_visible_child_name('first-run')
            self.first_run_view.connect('exit', self._on_first_run_exit)
            shared.schema.set_boolean('db-needs-update', False)
            return
            
        db_needs_update = shared.schema.get_boolean('db-needs-update')
        if db_needs_update:
            logging.info('Start db update')
            self.db_update_view = DbUpdateView()
            self._win_stack.add_named(child=self.db_update_view, name='db-update')
            self._win_stack.set_visible_child_name('db-update')
            self.db_update_view.connect('exit', self._on_db_update_exit)
            return
            
        self._win_stack.add_named(child=MainView(self), name='main')
        self._win_stack.set_visible_child_name('main')
        return

    def _on_network_changed(self, network_monitor: Gio.NetworkMonitor, network_available: bool) -> None:
        """
        Callback for "network-changed" signal.
        If no network is available, it turns on offline mode.

        Args:
            network_monitor (Gio.NetworkMonitor): the NetworkMonitor in use
            network_available (bool): whether or not the network is available

        Returns:
            None
        """

        logging.warning(f'Network changed, current status {network_available}')
        logging.warning(f'Offline mode: {not network_available}')

        shared.schema.set_boolean(
            'offline-mode', GLib.Variant.new_boolean(not network_available))

    def _on_first_run_exit(self, source: Gtk.Widget) -> None:
        """
        Callback for the "exit" signal. Changes the visible view.

        Args:
            None

        Returns:
            None
        """

        logging.info('First setup done')
        self._win_stack.add_named(child=MainView(self), name='main')
        self._win_stack.set_visible_child_name('main')

    def _on_db_update_exit(self, source: Gtk.Widget) -> None:
        """
        Callback for the "exit" of the db_update task. Changes the visible view.

        Args:
            None

        Returns:
            None
        """

        logging.info('Database Update done')
        self._win_stack.add_named(child=MainView(self), name='main')
        self._win_stack.set_visible_child_name('main')

    def _restore_state(self) -> None:
        """
        Restores the last known state of the window between runs.

        Args:
            None

        Returns:
            None
        """

        shared.schema.bind('win-width', self, 'default-width',
                           Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('win-height', self, 'default-height',
                           Gio.SettingsBindFlags.DEFAULT)
        shared.schema.bind('win-maximized', self, 'maximized',
                           Gio.SettingsBindFlags.DEFAULT)



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GObject, Gtk

from .. import shared  # type: ignore
from ..background_queue import BackgroundQueue


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/background_indicator.ui')
class BackgroundIndicator(Adw.Bin):
    """
    This class represents the indicator for background activities.

    Properties:
        queue (Gio.ListStore): the queue

    Methods:
        refresh(): updates the icon button

    Signals:
        None
    """

    __gtype_name__ = 'BackgroundIndicator'

    queue = GObject.Property(type=Gio.ListStore)

    _stack = Gtk.Template.Child()
    _model = Gtk.Template.Child()
    _image = Gtk.Template.Child()
    _spinner = Gtk.Template.Child()
    _list_view = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        self.queue = BackgroundQueue.get_queue()

        self._spinner.bind_property('visible', self._image, 'visible', GObject.BindingFlags.INVERT_BOOLEAN)
        self.queue.connect('notify::n-items', self._on_queue_change)

    def _on_queue_change(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for "notify::n-items" signal.
        Updates the model and refreshes the indicator.

        Args:
            pspec (GObject.ParamSpec): The GParamSpec of the property which changed
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        if self.queue.get_property('n-items') > 0:

            self._model.remove_all()
            for activity in self.queue:
                self._model.append(activity)

            self._stack.set_visible_child_name('filled')
            self.refresh()

    def refresh(self) -> None:
        """
        Checks the activities and shows a spinner as the button icon if at least one activity is running. If all activities are completed, the icon is set to a check mark.

        Args:
            None

        Returns:
            None
        """

        self._spinner.set_visible(not all(activity.completed for activity in self.queue))



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
from datetime import datetime, timedelta
from datetime import date
from gettext import gettext as _
from gettext import ngettext
from gettext import pgettext as C_
from typing import List, Tuple

from gi.repository import Adw, Gio, GObject, Gtk
from PIL import Image, ImageStat

from .. import shared  # type: ignore
from ..background_queue import (ActivityType, BackgroundActivity,
                                BackgroundQueue)
from ..dialogs.add_manual_dialog import AddManualDialog
from ..models.movie_model import MovieModel
from ..models.season_model import SeasonModel
from ..models.series_model import SeriesModel
from ..models.episode_model import EpisodeModel
from ..providers.local_provider import LocalProvider as local
from ..providers.tmdb_provider import TMDBProvider as tmdb
from ..widgets.episode_row import EpisodeRow
from ..widgets.theme_switcher import ThemeSwitcher


@Gtk.Template(resource_path=shared.PREFIX + '/ui/pages/details_page.ui')
class DetailsView(Adw.NavigationPage):
    """
    Widget that represents the details view.

    Properties:
        content (MovieModel or SeriesModel): content associated to the information shown

    Methods:
        None

    Signals:
        deleted: emited when the user confirms the deletion of the content
        edited: emited when a user confirms the edits for manually added content
    """

    __gtype_name__ = 'DetailsView'

    content = GObject.Property(type=object)

    __gsignals__ = {
        'deleted': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'edited': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    _view_stack = Gtk.Template.Child()
    _menu_btn = Gtk.Template.Child()
    _background_picture = Gtk.Template.Child()
    _poster_picture = Gtk.Template.Child()
    _title_lbl = Gtk.Template.Child()
    _tagline_lbl = Gtk.Template.Child()
    _genres_lbl = Gtk.Template.Child()
    _chip1_lbl = Gtk.Template.Child()
    _chip2_lbl = Gtk.Template.Child()
    _chip3_lbl = Gtk.Template.Child()
    _chip4_lbl = Gtk.Template.Child()
    _watched_btn = Gtk.Template.Child()
    _btn_content = Gtk.Template.Child()
    _edit_btn = Gtk.Template.Child()
    _update_btn = Gtk.Template.Child()
    _activate_notification_btn = Gtk.Template.Child()
    _notification_icon = Gtk.Template.Child()
    _description_box = Gtk.Template.Child()
    _overview_lbl = Gtk.Template.Child()
    _creator_box = Gtk.Template.Child()
    _creator_lbl = Gtk.Template.Child()
    _seasons_box = Gtk.Template.Child()
    _seasons_group = Gtk.Template.Child()
    _additional_info_box = Gtk.Template.Child()
    _flow_box = Gtk.Template.Child()
    _loading_lbl = Gtk.Template.Child()
    _notes_save_btn = Gtk.Template.Child()
    _notes_save_revealer = Gtk.Template.Child()
    _notes_textview = Gtk.Template.Child()

    def __init__(self, content: MovieModel | SeriesModel, content_view):
        super().__init__()

        self.content_view = content_view
        # Theme switcher (Adapted from https://gitlab.gnome.org/tijder/blueprintgtk/)
        themeswitcher = ThemeSwitcher()
        themeswitcher.connect(
            'themer-clicked', self._on_themeswitcher_clicked,)
        self._menu_btn.get_popover().add_child(themeswitcher, 'themeswitcher')

        if type(content) is MovieModel:
            self.content = local.get_movie_by_id(content.id)
        else:
            self.content = local.get_series_by_id(content.id)

        logging.info(
            f'Loading info [{"movie" if type(content) is MovieModel else "TV Serie"}] {self.content.title}') # type: ignore

        local.set_recent_change_status(self.content.id, False, type(    # type: ignore
            content) is MovieModel)  # reset recent_change since it was clicked on

        self.content_view.refresh_view()
        self.set_title(self.content.title)  # type: ignore
        self._view_stack.set_visible_child_name('loading')

        if shared.schema.get_int('win-width') >= 550:
            self.mobile = False
        else:
            self.mobile = True

        self._populate_data()

    @Gtk.Template.Callback()
    def _on_breakpoint_applied(self, breakpoint: Adw.Breakpoint) -> None:
        """
        Callback for the "applied" signal.
        Sets the orientation of the page based on the breakpoint.

        Args:
            breakpoint (Adw.Breakpoint): the breakpoint that was applied

        Returns:
            None
        """

        self.mobile = True
        if type(self.content) is SeriesModel:
            self._build_seasons_group()

    @Gtk.Template.Callback()
    def _on_breakpoint_unapplied(self, breakpoint: Adw.Breakpoint) -> None:
        """
        Callback for the "unapplied" signal.
        Sets the orientation of the page based on the breakpoint.

        Args:
            breakpoint (Adw.Breakpoint): the breakpoint that was unapplied

        Returns:
            None
        """

        self.mobile = False
        if type(self.content) is SeriesModel:
            self._build_seasons_group()

    def _populate_data(self) -> None:
        """
        Populates the widgets with the available information.

        Args:
            None

        Returns:
            None
        """

        # Both movies and tv series
        if self.content.backdrop_path:  # type: ignore

            if not Adw.StyleManager.get_default().get_high_contrast():
                self._background_picture.set_file(Gio.File.new_for_uri(
                    self.content.backdrop_path))  # type: ignore
                # type: ignore
                with Image.open(self.content.backdrop_path[7:]) as image:
                    stat = ImageStat.Stat(image.convert('L'))

                    luminance = [
                        min((stat.mean[0] + stat.extrema[0][0]) / 510, 0.7),
                        max((stat.mean[0] + stat.extrema[0][1]) / 510, 0.3),
                    ]
                self._background_picture.set_opacity(1 - luminance[0]
                                                     if Adw.StyleManager.get_default().get_dark()
                                                     else luminance[1])

        self._poster_picture.set_file(Gio.File.new_for_uri(
            self.content.poster_path))  # type: ignore

        self._title_lbl.set_text(self.content.title)  # type: ignore

        if self.content.tagline:  # type: ignore
            self._tagline_lbl.set_visible(True)
            self._tagline_lbl.set_text(self.content.tagline)  # type: ignore

        if self.content.genres:  # type: ignore
            self._genres_lbl.set_visible(True)
            self._genres_lbl.set_label(
                ', '.join(self.content.genres))  # type: ignore

        if self.content.release_date:  # type: ignore
            self._chip1_lbl.set_visible(True)
            self._chip1_lbl.set_text(date.fromisoformat(
                # type: ignore
                self.content.release_date).strftime('%d %b. %Y'))

        if self.content.manual:  # type: ignore
            self._edit_btn.set_visible(True)
        else:
            self._update_btn.set_visible(True)

        if self.content.overview:  # type: ignore
            self._description_box.set_visible(True)
            self._overview_lbl.set_label(self.content.overview)  # type: ignore

        # Movie specific
        if type(self.content) is MovieModel:
            self._watched_btn.set_visible(True)
            if self.content.watched:
                self._btn_content.set_label(_('Watched'))
                self._btn_content.set_icon_name('check-plain')
            else:
                self._btn_content.set_label(_('Mark as Watched'))
                self._btn_content.set_icon_name('watchlist')

            if self.content.runtime:
                self._chip2_lbl.set_visible(True)
                self._chip2_lbl.set_text(
                    self._format_runtime(self.content.runtime))
                self._chip2_lbl.set_tooltip_text(_("Runtime of Movie"))

            if not self.content.manual and datetime.strptime(self.content.release_date, '%Y-%m-%d') > datetime.now():
                self._notification_icon.set_visible(True)
                self._activate_notification_btn.set_visible(True)
                self._activate_notification_btn.set_active(
                    local.get_notification_list_status(self.content.id, movie=True))

        # TV series specific
        if type(self.content) is SeriesModel:
            if self.content.seasons_number:
                self._chip2_lbl.set_visible(True)
                # TRANSLATORS: {num} is the total number of seasons
                self._chip2_lbl.set_text(ngettext('{num} Season'.format(num=self.content.seasons_number),
                                                  '{num} Seasons'.format(
                                                      num=self.content.seasons_number),
                                         self.content.seasons_number))
                self._chip2_lbl.set_tooltip_text(_("Number of Seasons"))

            if self.content.episodes_number:
                self._chip3_lbl.set_visible(True)
                # TRANSLATORS: {num} is the total number of episodes
                self._chip3_lbl.set_text(ngettext('{num} Episode'.format(num=self.content.episodes_number),
                                                  '{num} Episodes'.format(
                                                      num=self.content.episodes_number),
                                                  self.content.episodes_number))

            if self.content.next_air_date:  # type: ignore
                self._chip4_lbl.set_visible(True)
                self._chip4_lbl.set_text(date.fromisoformat(
                    # type: ignore
                    self.content.next_air_date).strftime('%d %b. %Y'))

            if not self.content.manual and self.content.in_production:
                self._notification_icon.set_visible(True)
                self._activate_notification_btn.set_visible(True)
                self._activate_notification_btn.set_active(
                    local.get_notification_list_status(self.content.id))
            if self.content.created_by:
                self._creator_box.set_visible(True)
                self._creator_lbl.set_text(', '.join(self.content.created_by))

            self._seasons_box.set_visible(True)
            self._build_seasons_group()

        self._build_flow_box()
        self._notes_textview.get_buffer().set_text(self.content.notes)  # type: ignore
        self._notes_textview.get_buffer().connect(
            'changed', self._on_notes_textview_changed)

        self._view_stack.set_visible_child_name('filled')

    def _build_seasons_group(self) -> None:
        """
        Creates the widgets needed to show season and episodes metadata.

        Args:
            None

        Returns:
            None
        """

        list_box = self._seasons_group.get_first_child().get_last_child().get_first_child()
        list_box.remove_all()

        self._episode_rows = []

        for season in self.content.seasons:  # type: ignore
            season_row = Adw.ExpanderRow(title=season.title,
                                         subtitle=ngettext('{num} Episode'.format(num=season.episodes_number),
                                                           '{num} Episodes'.format(
                                                               num=season.episodes_number),
                                                           season.episodes_number))

            poster = Gtk.Picture(height_request=112,
                                 width_request=75,
                                 content_fit=Gtk.ContentFit.FILL,
                                 margin_top=12,
                                 margin_bottom=12)
            poster.add_css_class('still')
            poster.set_file(Gio.File.new_for_uri(season.poster_path))
            season_row.add_prefix(poster)

            button = Gtk.Button(valign=Gtk.Align.CENTER)

            if not self.mobile:
                btn_content = Adw.ButtonContent()

                if all(episode.watched for episode in season.episodes):
                    btn_content.set_label(_('Watched'))
                    btn_content.set_icon_name('check-plain')
                else:
                    btn_content.set_label(_('Mark as Watched'))
                    btn_content.set_icon_name('watchlist')

                button.set_child(btn_content)
            else:
                if all(episode.watched for episode in season.episodes):
                    button.set_icon_name('check-plain')
                else:
                    button.set_icon_name('watchlist')

            season_row.add_suffix(button)

            tmp = []
            for episode in season.episodes:
                episode_row = EpisodeRow(episode, small_controls=self.mobile)
                episode_row.connect(
                    'watched-clicked', self._on_episode_watch_clicked, (button, season, episode))
                episode_row.add_css_class("groupcolor")
                season_row.add_row(episode_row)
                tmp.append(episode_row)

            self._seasons_group.add(season_row)
            self._episode_rows.append((season, tmp))

            button.connect('clicked', self._on_season_watched_clicked,
                           (button, season, self._episode_rows))

    def _on_episode_watch_clicked(self,
                                  source: Gtk.Widget,
                                  data: Tuple[Gtk.Button, SeasonModel, EpisodeModel]) -> None:
        """
        Callback for "watched-clicked" signal.
        Called after an episode is (un)marked as watched, checks and updates, if needed, the watched button for the corresponding season.

        Args:
            source (Gtk.Widget): caller widget
            data(tuple[Gtk.Button, SeasonModel, EpisodeModel]): tuple with the Gtk.Button to change and the SeasonModel
                parent of the changed episode

        Returns:
            None
        """

        self.content = local.get_series_by_id(self.content.id)  # type: ignore

        season_idx = 0
        for idx, season in enumerate(self.content.seasons):  # type: ignore
            if season == data[1]:
                season_idx = idx

        # compare if the episode clicked was the newest released episode if this is the case set new_release to False
        if (season_idx+1) == int(self.content.last_episode_number.split('.')[0]) \
                and data[2].number == int(self.content.last_episode_number.split('.')[1]):
            local.set_new_release_status(self.content.id, False)

        if not self.mobile:
            btn_content = data[0].get_child()
            # type: ignore
            if all(episode.watched for episode in self.content.seasons[season_idx].episodes):
                btn_content.set_label(_('Watched'))
                btn_content.set_icon_name('check-plain')
            else:
                btn_content.set_label(_('Mark as Watched'))
                btn_content.set_icon_name('watchlist')
        else:
            # type: ignore
            if all(episode.watched for episode in self.content.seasons[season_idx].episodes):
                data[0].set_icon_name('check-plain')
            else:
                data[0].set_icon_name('watchlist')

        # Update season status
        local.mark_watched_series(self.content.id, all(  # type: ignore
            episode.watched for season in self.content.seasons for episode in season.episodes))  # type: ignore
        self.activate_action('win.refresh', None)

    def _on_season_watched_clicked(self,
                                   source: Gtk.Widget,
                                   data: Tuple[Gtk.Button, SeasonModel, List[Tuple[SeasonModel, List[EpisodeRow]]]]) -> None:
        """
        Callback for "clicked" signal.
        Marks a whole season as (un)watched, making changes in the db and updating the ui.

        Args:
            source (Gtk.Widget): caller widget
            data (Tuple[Gtk.Button, SeasonModel, List[Tuple[SeasonModel, List[EpisodeRow]]]]): tuple with the
                Gtk.Button to change, the SeasonModel of the modified season, and a list of tuples of all
                SeasonModels and associated EpisodeRows.

        Returns:
            None
        """

        self.content = local.get_series_by_id(self.content.id)  # type: ignore

        season_idx = 0
        for idx, season in enumerate(self.content.seasons):  # type: ignore
            if season == data[1]:
                season_idx = idx

        episode_rows = []
        for item in data[2]:
            if item[0] == self.content.seasons[season_idx]:  # type: ignore
                episode_rows = item[1]

        # determine if we want to set all episodes to watched or to not watched
        if type(data[0].get_child()) is Adw.ButtonContent:
            set_to_watched = data[0].get_child().get_icon_name() == 'watchlist'
        else:
            set_to_watched = data[0].get_icon_name() == 'watchlist'

        # Make changes in db
        for episode in self.content.seasons[season_idx].episodes:
            local.mark_watched_episode(
                episode.id, set_to_watched)  # type: ignore

        # if the season in which the last aired episode is in is clicked as watched remove new_release
        if season_idx == int(self.content.last_episode_number.split('.')[0]):
            local.set_new_release_status(self.content.id, False)

        # Update episode rows
        for episode_row in episode_rows:
            episode_row.set_watched_btn(set_to_watched)  # type: ignore

        # Update season expander
        if not self.mobile:
            btn_content = data[0].get_child()
            if set_to_watched:  # type: ignore
                btn_content.set_label(_('Watched'))
                btn_content.set_icon_name('check-plain')
            else:
                btn_content.set_label(_('Mark as Watched'))
                btn_content.set_icon_name('watchlist')
        else:
            if set_to_watched:  # type: ignore
                data[0].set_icon_name('check-plain')
            else:
                data[0].set_icon_name('watchlist')

        # refetch data from db since we have changed episode.watched earlier
        self.content = local.get_series_by_id(self.content.id)

        # Update season status
        local.mark_watched_series(self.content.id, all(
            episode.watched for season in self.content.seasons for episode in season.episodes))
        self.activate_action('win.refresh', None)

    def _build_flow_box(self) -> None:
        """
        Adds the available metadata to the 'additional information' section.

        Args:
            None

        Returns:
            None
        """

        if self._flow_box.get_child_at_index(0):
            return

        # Both movies and tv series
        if self.content.status:  # type: ignore
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=_('Status'))
            label.add_css_class('heading')
            box.append(label)
            box.append(Gtk.Label(label=self.content.status))  # type: ignore
            self._flow_box.append(box)

        if self.content.original_language and self.content.original_language.iso_name != 'xx':  # type: ignore
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=_('Original Language'))
            label.add_css_class('heading')
            box.append(label)
            # type: ignore
            box.append(
                Gtk.Label(label=self.content.original_language.name, lines=2, wrap=True))
            self._flow_box.append(box)

        if self.content.original_title:  # type: ignore
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=_('Original Title'))
            label.add_css_class('heading')
            box.append(label)
            # type: ignore
            box.append(
                Gtk.Label(label=self.content.original_title, lines=2, wrap=True))
            self._flow_box.append(box)

        # Movie specific
        if type(self.content) is MovieModel:
            if self.content.budget:
                box = Gtk.Box(
                    orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
                label = Gtk.Label(label=_('Budget'))
                label.add_css_class('heading')
                box.append(label)
                box.append(Gtk.Label(label='US$ {budget:0.0f}'.format(
                    budget=self.content.budget)))
                self._flow_box.append(box)

            if self.content.revenue:
                box = Gtk.Box(
                    orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
                label = Gtk.Label(label=_('Revenue'))
                label.add_css_class('heading')
                box.append(label)
                box.append(Gtk.Label(label='US$ {revenue:0.0f}'.format(
                    revenue=self.content.revenue)))
                self._flow_box.append(box)

        # TV series specific
        if type(self.content) is SeriesModel:
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=_('In Production'))
            label.add_css_class('heading')
            box.append(label)
            box.append(Gtk.Label(label=_('Yes')
                       if self.content.in_production else _('No')))
            self._flow_box.append(box)

        if self.content.id and self.content.manual == False:
            box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
            label = Gtk.Label(label=_('TMDB ID'))
            label.add_css_class('heading')
            box.append(label)
            url_str = '<a href="https://www.themoviedb.org/{category}/{id}" title="TMDB Webpage"> {id} </a>' \
                .format(category=("tv") if type(self.content) is SeriesModel else ("movie"), id=self.content.id)
            box.append(Gtk.Label(label=url_str, use_markup=True))
            self._flow_box.append(box)

        if self._flow_box.get_child_at_index(0) is None:
            self._additional_info_box.set_visible(False)

    def _format_runtime(self, runtime: str) -> str:
        """
        Formats the runtime in hours and minutes.

        Args:
            runtime (str): runtime in minutes

        Returns:
            formatted duration string in hours and minutes
        """

        h, m = divmod(int(runtime), 60)

        if h > 0:
            # TRANSLATORS: {h} and {m} are the runtime hours and minutes respectively
            return _('{h}h {m}min').format(h=h, m=m)
        else:
            # TRANSLATORS: {m} is the runtime minutes
            return _('{m}min').format(m=m)

    @Gtk.Template.Callback('_on_watched_btn_clicked')
    def _on_watched_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Updates the watched flag for the content in the db and changes the button text accordingly.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        if type(self.content) is MovieModel:
            local.mark_watched_movie(self.content.id, not self.content.watched)

        self.content.watched = not self.content.watched  # type: ignore

        if self.content.watched:
            self._btn_content.set_label(_('Watched'))
            self._btn_content.set_icon_name('check-plain')
        else:
            self._btn_content.set_label(_('Mark as Watched'))
            self._btn_content.set_icon_name('watchlist')

        self.activate_action('win.refresh', None)

    @Gtk.Template.Callback('_on_edit_btn_clicked')
    def _on_edit_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Opens the edit window to change the metadata of the content. Available only for manually added content.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        logging.info(
            f'Editing [{"movie" if type(self.content) is MovieModel else "TV Serie"}] {self.content.title}')

        dialog = AddManualDialog(edit_mode=True, content=self.content)
        dialog.connect('edit-saved', self._on_edit_saved)
        dialog.present(self)

    def _on_edit_saved(self, source: Gtk.Widget, content: MovieModel | SeriesModel) -> None:
        """
        Callback for "edit-saved" signal.
        Replaces the navigation stack with an updated top page resulting in a content refresh.

        Args:
            source (Gtk.Widget): caller widget
            content (MovieModel or SeriesModel): updated content to show

        Returns:
            None
        """

        root_page = self.get_ancestor(
            Adw.NavigationView).get_previous_page(self)
        self.get_ancestor(Adw.NavigationView).replace(
            [root_page, DetailsView(content, self.content_view)])

    @Gtk.Template.Callback('_on_update_btn_clicked')
    def _on_update_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Adds a background activity to start a manual update.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._view_stack.set_visible_child_name('loading')

        # TRANSLATORS: {title} is the showed content's title
        self._loading_lbl.set_label(_('Updating {title}').format(
            title=self.content.title))  # type: ignore

        BackgroundQueue.add(
            activity=BackgroundActivity(
                activity_type=ActivityType.UPDATE,
                # TRANSLATORS: {title} is the content's title
                title=_('Update {title}').format(title=self.content.title),
                task_function=self._update),
            on_done=self._on_update_done)

    @Gtk.Template.Callback('_activate_notification_btn_toggled')
    def _activate_notification_btn_toggled(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Adds the series to the notification list
        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """
        movie = type(self.content) == MovieModel
        local.set_notification_list_status(
            self.content.id, self._activate_notification_btn.get_active(), movie=movie)
        # if we remove the content from the notification_list then remove the new/soon_release flags and refresh the ContentView
        if not self._activate_notification_btn.get_active():
            local.set_new_release_status(self.content.id, False, movie=movie)
            local.set_soon_release_status(self.content.id, False, movie=movie)
            self.content_view.refresh_view()
        else:  # if we add content we check if should set the soon_release flag
            compare_date = self.content.release_date if movie else self.content.next_air_date
            # TODO make this a variable and sync with main_view.py
            if len(compare_date) > 0 and datetime.strptime(compare_date, '%Y-%m-%d') < datetime.now() + timedelta(days=14 if movie else 7):
                local.set_soon_release_status(
                    self.content.id, True, movie=movie)
                self.content_view.refresh_view()

    def _update(self, activity: BackgroundActivity) -> None:
        """
        Fetches updated information and updates the stored copy. Additionally it handles ui updates.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """

        if type(self.content) is MovieModel:
            self.new_content = MovieModel(tmdb.get_movie(self.content.id))
            local.update_movie(old=self.content, new=self.new_content)
            self.new_content = local.get_movie_by_id(self.content.id)
        else:
            self.new_content = SeriesModel(tmdb.get_serie(self.content.id))
            local.update_series(old=self.content, new=self.new_content)
            self.new_content = local.get_series_by_id(self.content.id)

    def _on_update_done(self,
                        source: GObject.Object,
                        result: Gio.AsyncResult,
                        cancellable: Gio.Cancellable,
                        activity: BackgroundActivity):
        """Callback to complete async activity"""

        root_page = self.get_ancestor(
            Adw.NavigationView).get_previous_page(self)
        self.get_ancestor(Adw.NavigationView).replace(
            [root_page, DetailsView(self.new_content, self.content_view)])
        self._loading_lbl.set_label(_('Loading Metadata…'))
        self._view_stack.set_visible_child_name('filled')
        activity.end()

    @Gtk.Template.Callback('_on_delete_btn_clicked')
    def _on_delete_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Asks the user for a confirmation after a delete request.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        logging.debug('Show delete dialog')

        # TRANSLATORS: {title} is the content's title
        dialog = Adw.AlertDialog.new(heading=C_('message dialog heading', 'Delete {title}?').format(
            title=self.content.title),  # type: ignore
            body=C_('message dialog body', 'This title will be deleted from your watchlist.'))
        dialog.add_response('cancel', C_('message dialog action', '_Cancel'))
        dialog.add_response('delete', C_('message dialog action', '_Delete'))
        dialog.set_default_response('delete')
        dialog.set_close_response('cancel')
        dialog.set_response_appearance(
            'delete', Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.choose(self, None, self._on_message_dialog_choose, None)

    def _on_message_dialog_choose(self, source: GObject.Object | None, result: Gio.AsyncResult, user_data: object | None) -> None:
        """
        Callback for the message dialog.
        Finishes the async operation and retrieves the user response. If the later is positive, adds a background activity to delete the content.

        Args:
            source (Gtk.Widget): object that started the async operation
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        result = Adw.AlertDialog.choose_finish(source, result)
        if result == 'cancel':
            logging.debug('Delete dialog: cancel, aborting')
            return

        self.get_ancestor(Adw.NavigationView).pop()
        logging.debug(f'Delete dialog: confim, delete {self.content.title}') # type: ignore
        BackgroundQueue.add(
            activity=BackgroundActivity(
                activity_type=ActivityType.REMOVE,
                # TRANSLATORS: {title} is the content's title
                title=_('Delete {title}').format(
                    title=self.content.title),  # type: ignore
                task_function=self._delete),
            on_done=self._on_delete_done)

    def _delete(self, activity: BackgroundActivity) -> None:
        """
        Deletes the content from the db.

        Args:
            activity (BackgroundActivity): the calling activity

        Returns:
            None
        """

        if type(self.content) is MovieModel:
            local.delete_movie(self.content.id)
        else:
            local.delete_series(self.content.id)  # type: ignore

    def _on_delete_done(self,
                        source: GObject.Object,
                        result: Gio.AsyncResult,
                        cancellable: Gio.Cancellable,
                        activity: BackgroundActivity):
        """Callback to complete async activity"""

        self.emit('deleted')
        activity.end()

    def _on_themeswitcher_clicked(self,
                                  source: Gtk.Widget) -> None:
        """
        Callback for "themer-clicked" signal.
        Called after an theme is switched changes background picture if needed

        Args:
            source (Gtk.Widget): caller widget

        Returns:
            None
        """

        if self.content.backdrop_path:  # type: ignore
            if not Adw.StyleManager.get_default().get_high_contrast():
                self._background_picture.set_file(Gio.File.new_for_uri(
                    self.content.backdrop_path))  # type: ignore
                # type: ignore
                with Image.open(self.content.backdrop_path[7:]) as image:
                    stat = ImageStat.Stat(image.convert('L'))

                    luminance = [
                        min((stat.mean[0] + stat.extrema[0][0]) / 510, 0.7),
                        max((stat.mean[0] + stat.extrema[0][1]) / 510, 0.3),
                    ]
                self._background_picture.set_opacity(1 - luminance[0]
                                                     if Adw.StyleManager.get_default().get_dark()
                                                     else luminance[1])

    @Gtk.Template.Callback()
    def _on_notes_btn_clicked(self, user_data: GObject.GPointer | None) -> None:
        """
        Callback for the "clicked" signal.
        Saves the notes for the content in the db.

        Args:
            user_data (GObject.GPointer or None): additional data passed to the callback

        Returns:
            None
        """

        notes = self._notes_textview.get_buffer().get_text(
            self._notes_textview.get_buffer().get_start_iter(),
            self._notes_textview.get_buffer().get_end_iter(),
            True
        )

        if type(self.content) is MovieModel:
            local.update_movie_notes(self.content.id, notes)  # type: ignore

        if type(self.content) is SeriesModel:
            local.update_serie_notes(self.content.id, notes)  # type: ignore

        self._notes_save_revealer.set_reveal_child(False)
        self.content.notes = notes  # type: ignore
        self.activate_action('win.refresh', None)

    def _on_notes_textview_changed(self, user_data: GObject.GPointer | None) -> None:
        """
        Callback for the "changed" signal.
        Shows the save button.

        Args:
            user_data (GObject.GPointer or None): additional data passed to the callback

        Returns:
            None
        """

        self._notes_save_revealer.set_reveal_child(True)



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, GObject, Gtk

from .. import shared  # type: ignore
from ..models.episode_model import EpisodeModel


@Gtk.Template(resource_path=shared.PREFIX + '/ui/pages/edit_episode_page.ui')
class EditEpisodeNavigationPage(Adw.NavigationPage):
    """
    This class represents the 'edit episode' NavigationPane.

    Properties:
        None

    Methods:
        None

    Signals:
        edit-saved (title: str, episode_number: int, runtime: int, overview: str, still_uri: str): emited when the user clicks the save button
    """

    __gtype_name__ = 'EditEpisodeNavigationPage'

    _still = Gtk.Template.Child()
    _episode_spin_row = Gtk.Template.Child()
    _title_entry = Gtk.Template.Child()
    _runtime_spin_row = Gtk.Template.Child()
    _overview_text = Gtk.Template.Child()

    __gsignals__ = {
        'edit-saved': (GObject.SIGNAL_RUN_LAST, None, (str, int, int, str, str, bool,)),
    }

    def __init__(self,
                 title: str = '',
                 episode_number: int = 0,
                 runtime: int = 0,
                 overview: str = '',
                 still_uri: str = f'resource://{shared.PREFIX}/blank_still.jpg',
                 watched: bool = False):

        super().__init__()

        self._title = title
        self._episode_number = episode_number
        self._runtime = runtime
        self._overview = overview
        self._still_uri = still_uri
        self._watched = watched

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Sets the fields/still image to the provided values

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._overview_text.remove_css_class('view')

        self._title_entry.set_text(self._title)
        self._runtime_spin_row.set_value(self._runtime)
        self._still.set_blank_image(f'resource://{shared.PREFIX}/blank_still.jpg')
        if self._still_uri.startswith('file'):
            self._still.set_image(self._still_uri)
        self._overview_text.get_buffer().set_text(self._overview, -1)
        self._episode_spin_row.set_value(self._episode_number)

        self._title_entry.grab_focus()

    @Gtk.Template.Callback('_enable_save')
    def _enable_save(self, source: Gtk.Widget, title: str, episode_number: int) -> bool:
        """
        Closure to determine if the 'save' button should be enabled or not.

        Args:
            source (Gtk.Widget): caller widget
            title (str): title text
            episode_number (int): episode number

        Returns:
            bool
        """

        return True if title and episode_number > 0 else False

    @Gtk.Template.Callback('_on_save_btn_clicked')
    def _on_save_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for the "clicked" signal.
        Emits the "edit-saved" signal and pops the NavigationPage.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        buffer = self._overview_text.get_buffer()
        start_iter = buffer.get_start_iter()
        end_iter = buffer.get_end_iter()

        overview = buffer.get_text(start_iter, end_iter, False)
        title = self._title_entry.get_text()
        episode_number = int(self._episode_spin_row.get_value())
        runtime = int(self._runtime_spin_row.get_value())
        still_uri = self._still.get_uri()

        self.emit('edit-saved', title, episode_number, runtime, overview, still_uri, self._watched)
        self.get_ancestor(Adw.NavigationView).pop()



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gio, GLib, GObject, Gtk

from .. import shared  # type: ignore


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/image_selector.ui')
class ImageSelector(Adw.Bin):
    """
    This class represents the image selector and previewer with options to open a file and, if one is already opened,
    to delete the selection.

    Properties:
        content-fit (Gtk.ContentFit): content fit for the image

    Methods:
        set_file(file: Gio.File): sets the shown file
        get_uri(): gets the uri for the shown file

    Signals:
        None
    """

    __gtype_name__ = 'ImageSelector'

    content_fit = GObject.Property(type=Gtk.ContentFit, default=Gtk.ContentFit.FILL)
    shown_image = GObject.Property(type=str, default=f'resource://{shared.PREFIX}/blank_poster.jpg')
    blank_image = GObject.Property(type=str, default=f'resource://{shared.PREFIX}/blank_poster.jpg')

    _poster_picture = Gtk.Template.Child()
    _edit_btn = Gtk.Template.Child()
    _spinner = Gtk.Template.Child()
    _delete_revealer = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data):
        self._poster_picture.set_file(Gio.File.new_for_uri(self.shown_image))

    @Gtk.Template.Callback('_on_edit_btn_clicked')
    def _on_edit_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Setups and shows a file chooser dialog to choose a new image.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._edit_btn.set_sensitive(False)
        self._spinner.set_visible(True)

        file_filter_store = Gio.ListStore.new(Gtk.FileFilter)
        file_filter = Gtk.FileFilter()
        file_filter.add_pixbuf_formats()
        file_filter_store.append(file_filter)

        self.dialog = Gtk.FileDialog.new()
        self.dialog.set_modal(True)
        self.dialog.set_filters(file_filter_store)
        self.dialog.open(self.get_ancestor(Gtk.Window), None, self._on_file_open_complete, None)

    def _on_file_open_complete(self,
                               source: Gtk.Widget,
                               result: Gio.AsyncResult,
                               user_data: object | None) -> None:
        """
        Callback for the file dialog.
        Finishes the file selection and, if successfull, shows the new selected image.

        Args:
            source (Gtk.Widget): caller widget
            result (Gio.AsyncResult): a Gio.AsyncResult
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        try:
            poster_file = self.dialog.open_finish(result)
        except GLib.GError:
            poster_file = None

        if poster_file:
            self.shown_image = poster_file.get_uri()
            self._poster_picture.set_file(poster_file)
            self._delete_revealer.set_reveal_child(True)

        self._spinner.set_visible(False)
        self._edit_btn.set_sensitive(True)

    @Gtk.Template.Callback('_on_delete_btn_clicked')
    def _on_delete_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Restores the blank image and hides the delete button.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self.shown_image = self.blank_image
        self._poster_picture.set_file(Gio.File.new_for_uri(self.shown_image))
        self._delete_revealer.set_reveal_child(False)

    def set_blank_image(self, image_uri: str) -> None:
        """
        Sets the blank image and shows it.

        Args:
            image_uri (str): uri to use

        Returns:
            None
        """

        self.blank_image = image_uri
        self.shown_image = self.blank_image
        self._poster_picture.set_file(Gio.File.new_for_uri(self.shown_image))

    def set_image(self, image_uri: str) -> None:
        """
        Sets the image.

        Args:
            image_uri (str): uri to use

        Returns:
            None
        """

        self.shown_image = image_uri
        self._poster_picture.set_file(Gio.File.new_for_uri(self.shown_image))
        self._delete_revealer.set_reveal_child(True)

    def get_uri(self) -> str:
        """
        Returns the shown image uri.

        Args:
            None

        Returns:
            string with the uri
        """

        return self._poster_picture.get_file().get_uri()



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gettext import gettext as _
from typing import List

from gi.repository import Adw, GObject, Gtk

from .. import shared  # type: ignore
from ..pages.edit_episode_page import EditEpisodeNavigationPage
from ..widgets.episode_row import EpisodeRow


@Gtk.Template(resource_path=shared.PREFIX + '/ui/dialogs/edit_season.ui')
class EditSeasonDialog(Adw.Dialog):
    """
    This class represents the window to edit a season.

    Properties:
        None

    Methods:
        update_episodes_ui(): rebuilds the ui to reflect changes to episodes

    Signals:
        edit-saved (title: str, poster_uri: str, episodes: List[tuple]): emited when the user clicks the save button
    """

    __gtype_name__ = 'EditSeasonDialog'

    _navigation_view = Gtk.Template.Child()
    _save_btn = Gtk.Template.Child()
    _poster = Gtk.Template.Child()
    _title_entry = Gtk.Template.Child()
    _episodes_group = Gtk.Template.Child()

    __gsignals__ = {
        'edit-saved': (GObject.SIGNAL_RUN_FIRST, None, (str, str, object,)),
    }

    def __init__(self,
                 parent: Gtk.Window,
                 title: str = _('Season'),
                 poster_uri: str = f'resource://{shared.PREFIX}/blank_poster.jpg',
                 episodes: List[tuple] | None = None):

        super().__init__()
        self.parent = parent

        self._title = title
        self._poster_uri = poster_uri
        self._episodes = episodes if episodes else []

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Sets the title text and poster image, populates the episode data.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._title_entry.set_text(self._title)
        if self._poster_uri.startswith('file'):
            self._poster.set_image(self._poster_uri)

        self._title_entry.grab_focus()

        for episode in self._episodes:
            self._episodes_group.add(EpisodeRow(title=episode[0],
                                                episode_number=episode[1],
                                                runtime=episode[2],
                                                overview=episode[3],
                                                still_uri=episode[4],
                                                watched=episode[5],
                                                editable=True)
                                     )

    @Gtk.Template.Callback('_on_title_entry_changed')
    def _on_title_entry_changed(self, user_data: object | None) -> None:
        """
        Callback for "changed" signal.
        Wrapper around self._enable_save_btn().

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self._enable_save_btn()

    def _enable_save_btn(self) -> None:
        """
        Checks whether the "save" button should be made active or not.

        Args:
            None

        Returns:
            None
        """

        if self._title_entry.get_text() and len(self._episodes) > 0:
            self._save_btn.set_sensitive(True)
            return

        self._save_btn.set_sensitive(False)

    @Gtk.Template.Callback('_on_save_btn_clicked')
    def _on_save_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Emits the "edit-saved" signal and closes the window.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        self.emit('edit-saved', self._title_entry.get_text(), self._poster.get_uri(), self._episodes)
        self.close()

    @Gtk.Template.Callback('_on_add_btn_clicked')
    def _on_add_btn_clicked(self, user_data: object | None) -> None:
        """
        Callback for "clicked" signal.
        Shows the "edit episode" page.

        Args:
            user_data (object or None): user data passed to the callback.

        Returns:
            None
        """

        edit_episode_page = EditEpisodeNavigationPage(episode_number=len(self._episodes)+1)
        edit_episode_page.connect('edit-saved', self._on_episode_saved)
        self._navigation_view.push(edit_episode_page)

    def _on_episode_saved(self,
                          source: Gtk.Widget,
                          title: str,
                          episode_number: int,
                          runtime: int,
                          overview: str,
                          still_uri: str,
                          watched: bool) -> None:
        """
        Callback for "edit-saved" signal.
        Appends the recieved data as a tuple in the episodes list and updates the ui.

        Args:
            source (Gtk.Widget): caller widget
            title (str): epiosode title
            episode_number (int): episode number
            runtime (int): episode runtime in minutes
            overview (str): episode overview
            still_uri (str): episode still uri

        Returns:
            None
        """

        self._episodes.append((title, episode_number, runtime, overview, still_uri, watched))
        self.update_episodes_ui()

    def update_episodes_ui(self) -> None:
        """
        Rebuilds the ui to reflect changes to episodes.

        Args:
            None

        Returns:
            None
        """

        # Empty PreferencesGroup
        list_box = self._episodes_group.get_first_child().get_last_child().get_first_child()  # ugly workaround
        list_box.remove_all()

        # Fill PreferencesGroup
        for episode in self._episodes:
            self._episodes_group.add(EpisodeRow(title=episode[0],
                                                episode_number=episode[1],
                                                runtime=episode[2],
                                                overview=episode[3],
                                                still_uri=episode[4],
                                                watched=episode[5],
                                                editable=True))
        self._enable_save_btn()

    def get_episode(self,
                    title: str,
                    episode_number: int,
                    runtime: int,
                    overview: str,
                    still_uri: str) -> tuple:
        """
        Compares and retrieves the tuple containing the passed data.

        Args:
            title (str): a title
            episode_number (int): an episode number
            runtime (int): a runtime in minutes
            overview (str): an overview
            still_uri (str): an uri

        Returns:
            tuple matching the passed data
        """

        for episode in self._episodes:
            if (episode[0] == title and
                episode[1] == episode_number and
                episode[2] == runtime and
                episode[3] == overview and
                    episode[4] == still_uri):
                return episode
        return ()



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, GObject, Gtk

from .. import shared  # type: ignore


@Gtk.Template(resource_path=shared.PREFIX + '/ui/widgets/background_activity_row.ui')
class BackgroundActivityRow(Adw.Bin):
    """
    This class represents a row in the BackgroundIndicator popover.

    Properties:
        title (str): a title
        activity_type (str): an activity type, name as in ActivityType
        completed (bool): indicates if the activity is completed

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'BackgroundActivityRow'

    title = GObject.Property(type=str, default='')
    activity_type = GObject.Property(type=str, default='')
    completed = GObject.Property(type=bool, default=False)
    has_error = GObject.Property(type=bool, default=False)

    _icon = Gtk.Template.Child()
    _progress_bar = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        self.connect('notify::completed', self._on_complete)

    @Gtk.Template.Callback('_on_map')
    def _on_map(self, user_data: object | None) -> None:
        """
        Callback for "map" signal.
        Sets the icon based on the completion status and activity type, and starts the progress bar.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        if not self.completed:
            match self.activity_type:
                case 'ADD':
                    self._icon.set_from_icon_name('plus')
                case 'REMOVE':
                    self._icon.set_from_icon_name('user-trash-symbolic')
                case 'UPDATE':
                    self._icon.set_from_icon_name('update')
            GObject.timeout_add(500, self._on_timeout, None)
        else:
            if self.has_error:
                self._progress_bar.add_css_class('progress_error')
                self._icon.set_from_icon_name('warning')
            else:
                self._progress_bar.add_css_class('progress_complete')
                self._icon.set_from_icon_name('check-plain')
            self._progress_bar.set_fraction(1)

    def _on_timeout(self, user_data: object | None) -> bool:
        """
        Callback for GObject.timeout_add.
        Pulses the progress bar.

        Args:
            user_data (object or None): additional data passed to the callback

        Returns:
            True if the timeout should be called again, False otherwise
        """

        if not self.completed:
            self._progress_bar.pulse()
            return True
        else:
            return False

    def _on_complete(self, pspec: GObject.ParamSpec, user_data: object | None) -> None:
        """
        Callback for "notify::completed" signal.
        Sets the icon and progres bar to show a completed status, updates the background indicator.

        Args:
            pspec (GObject.ParamSpec): The GParamSpec of the property which changed
            user_data (object or None): additional data passed to the callback

        Returns:
            None
        """

        self._icon.set_from_icon_name('check-plain')
        self._progress_bar.set_fraction(1)

        if self.get_ancestor(Adw.ApplicationWindow):
            self.get_ancestor(Adw.ApplicationWindow).activate_action(
                'win.update-backgroud-indicator')



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import glob
import os
import re
from typing import List

import requests
from gi.repository import GObject

import src.providers.local_provider as local
import src.providers.tmdb_provider as tmdb

from .. import shared  # type: ignore
from ..models.episode_model import EpisodeModel


class SeasonModel(GObject.GObject):
    """
    This class represents a season object stored in the db.

    Properties:
        episodes (List[EpisodeModel]): list of episodes in self
        episodes_number (int): number of episodes in self
        id (str): season id
        number (int): season number
        overview (str): season overview
        poster_path (str): uri of the season's poster
        show_id (int):  id of the show the searies belongs to
        title (str): season title

    Methods:
        None

    Signals:
        None
    """

    __gtype_name__ = 'SeasonModel'

    episodes = GObject.Property(type=object)
    episodes_number = GObject.Property(type=int, default=0)
    id = GObject.Property(type=str, default='')
    number = GObject.Property(type=int, default=0)
    overview = GObject.Property(type=str, default='')
    poster_path = GObject.Property(type=str, default='')
    show_id = GObject.Property(type=str, default='')
    title = GObject.Property(type=str, default='')

    def __eq__(self, other) -> bool:
        """
        Custom comparing fuction, overrides '==' operator.

        Args:
            other: object to compare to

        Returns:
            bool result of the operation
        """

        if type(other) is not SeasonModel:
            return False

        if (self.episodes_number == other.episodes_number and
            self.id == other.id and
            self.number == other.number and
            self.overview == other.overview and
            self.poster_path == other.poster_path and
            self.show_id == other.show_id and
                self.title == other.title):
            return True
        else:
            return False

    def __init__(self, show_id: int = 0, d=None, t=None):
        super().__init__()

        if d is not None:
            self.episodes_number = d['episode_count']
            self.id = d['id']
            self.number = d['season_number']
            self.overview = re.sub(r'\s{2}', ' ', d['overview'])
            self.poster_path = self._download_poster(show_id, d['poster_path'])
            self.title = d['name']
            self.show_id = show_id

            self.episodes = self._parse_episodes(
                tmdb.TMDBProvider.get_season_episodes(show_id, self.number))
        else:
            self.episodes_number = t[0]  # type: ignore
            self.id = t[1]  # type: ignore
            self.number = t[2]  # type: ignore
            self.overview = t[3]  # type: ignore
            self.poster_path = t[4]  # type: ignore
            self.title = t[5]  # type: ignore
            self.show_id = t[6]  # type: ignore

            if len(t) == 8:  # type: ignore
                self.episodes = t[7]    # type: ignore
            else:
                self.episodes = local.LocalProvider.get_season_episodes(
                    self.show_id, self.number)  # type: ignore

    def _download_poster(self, show_id: int, path: str) -> str:
        """
        Returns the uri of the poster image on the local filesystem, downloading if necessary.

        Args:
            path (str): path to dowload from

        Returns:
            str with the uri of the poster image
        """

        if not path:
            return f'resource://{shared.PREFIX}/blank_poster.jpg'

        if not os.path.exists(f'{shared.series_dir}/{show_id}/{self.number}'):
            os.makedirs(f'{shared.series_dir}/{show_id}/{self.number}')

        files = glob.glob(
            f'{path[1:-4]}.jpg', root_dir=f'{shared.series_dir}/{show_id}/{self.number}')
        if files:
            return f'file://{shared.series_dir}/{show_id}/{self.number}/{files[0]}'

        url = f'https://image.tmdb.org/t/p/w500{path}'
        try:
            r = requests.get(url)
            if r.status_code == 200:
                with open(f'{shared.series_dir}/{show_id}/{self.number}{path}', 'wb') as f:
                    f.write(r.content)
                return f'file://{shared.series_dir}/{show_id}/{self.number}{path}'
            else:
                return f'resource://{shared.PREFIX}/blank_poster.jpg'
        except (requests.exceptions.ConnectionError, requests.exceptions.SSLError):
            return f'resource://{shared.PREFIX}/blank_poster.jpg'

    def _parse_episodes(self, episodes: dict) -> List[EpisodeModel]:
        """
        Parses episode data comming from tmdb into a list of EpisodeModels.

        Args:
            episodes (dict): dict from the api

        Returns:
            List of EpisodeModels
        """

        episode_list = []

        for episode in episodes:
            episode_list.append(EpisodeModel(d=episode))
        return episode_list



# Copyright (C) 2023 Alessandro Iepure
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import shutil
import sqlite3
import zipfile
import shutil
import json
from typing import List, Dict, Any
from pathlib import Path
from PIL import Image, ImageStat
from datetime import datetime, timedelta

from .. import shared  # type: ignore
from ..models.episode_model import EpisodeModel
from ..models.language_model import LanguageModel
from ..models.movie_model import MovieModel
from ..models.season_model import SeasonModel
from ..models.series_model import SeriesModel
from ..providers.tmdb_provider import TMDBProvider as tmdb


class LocalProvider:
    """
    This class provides methods to interface with the local db.

    Properties:
        None

    Methods:
        create_movies_table(): Creates the table used to store movie details in a local database
        create_series_table(): Creates the table used to store tv series details in a local database
        create_languages_table(): Creates the table used to store the available languages in a local database
        create_tables(): Convenience method to create all tables with a single call
        update_series_table(): Checks for missing and adds new columns to the database
        add_language(language: LanguageModel): Inserts the provided LanguageModel in the languages table
        add_movie(id: int, movie: MovieModel): Inserts a movie in the movies table, querying the data from TMDB if only
            id is provided.
        add_series(id: int, serie: SeriesModel): Inserts a tv series in the series table, querying the data from TMDB
            if only id is provided.
        add_content(id: int, media_type: str): Convenience method to add movies and series from TMDB without using
            separate methods
        get_language_by_code(iso_code: str): Retrieves a language from the db via its iso_639_1 code.
        get_movie_by_id(id: str): Retrieves a movie from the db via its id
        get_all_movies(): Retrieves all movies from the db
        mark_watched_movie(id: str, status: bool): Sets the watched flag on the movie with the provided id.
        delete_movie(id: str): Deletes the movie with the provided id, removing associated files too.
        get_all_seasons(show: str): Retrieves metadata for all seasons of a show.
        get_season_episodes(show: str, season_number: int): Retrieves the episodes for a season of a show
        get_series_by_id(id: str): Retrieves the series with the provided id.
        get_all_series(): Retrieves all tv series from the db.
        mark_watched_series(id: str, watched: bool): Sets the watched flag on all episodes in the series and the series
            itself.
        delete_series(id: str): Deletes the tv series with the provided id, removing associated files too.
        get_all_languages(): Retrieves all languages from the db.
        get_next_manual_movie(): Calculates the next id for a manually added movie.
        get_next_manual_series(): Calculates the next id for a manually added tv series.
        get_next_manual_season(): Calculates the next id for a manually added season.
        get_next_manual_episode(): Calculates the next id for a manually added episode.
        get_language_by_name(name: str): Retrieves a language from the db via its name.
        update_movie(old: MovieModel, new: MovieModel): Updates a movie with new data.
        mark_watched_episode(id: str, watched: bool): Sets the watched flag on the specified episode.
        get_episode_by_id(id: str): Retrieves an episode from the db via its id.
        set_notification_list_status(id: int, value: bool):  Sets the activate_notification field of the given content to value
        get_notification_list_status(id: int): Returns if the content given by the id is on the notification list
        set_new_release_status(id: int, value: bool): Sets the new_release field of the given content to value
        get_new_release_status(id: int): Returns if the content given by the id has had a new release
        set_soon_release_status(id: int, value: bool): Sets the soon_release field of the given content to value
        get_soon_release_status(id: int): Returns if the content given by the id has a new release soon
        export_data(path: Path): Exports the database a zip archive to the given path
    """

    @staticmethod
    def create_movies_table() -> None:
        """
        Creates the table used to store movie details in a local database.

        Args:
            None

        Returns:
            None
        """

        with sqlite3.connect(shared.db) as connection:
            logging.debug('[db] Create movie table')
            sql = """CREATE TABLE IF NOT EXISTS movies (
                        activate_notification BOOLEAN,
                        add_date TEXT,
                        backdrop_path TEXT,
                        budget INTEGER,
                        color BOOLEAN,
                        genres TEXT,
                        id TEXT PRIMARY KEY,
                        manual BOOLEAN,
                        notes TEXT,
                        new_release BOOLEAN,
                        original_language TEXT,
                        original_title TEXT,
                        overview TEXT,
                        poster_path TEXT,
                        recent_change BOOLEAN,
                        release_date TEXT,
                        revenue INTEGER,
                        runtime INTEGER,
                        soon_release BOOLEAN,
                        status TEXT,
                        tagline TEXT,
                        title TEXT,
                        watched BOOLEAN,
                        FOREIGN KEY (original_language) REFERENCES languages (iso_639_1)
                     );"""
            connection.cursor().execute(sql)
            connection.commit()

    @staticmethod
    def create_series_table() -> None:
        """
        Creates the table used to store tv series details in a local database.

        Args:
            None

        Returns:
            None
        """

        with sqlite3.connect(shared.db) as connection:
            logging.debug('[db] Create series, seasons, and episodes tables')
            series_sql = """CREATE TABLE IF NOT EXISTS series (
                            add_date TEXT,
                            backdrop_path TEXT,
                            color BOOLEAN,
                            created_by TEXT,
                            episodes_number INT,
                            genres TEXT,
                            id TEXT PRIMARY KEY,
                            in_production BOOLEAN,
                            last_air_date TEXT,
                            last_episode_number TEXT,
                            manual BOOLEAN,
                            next_air_date TEXT,
                            new_release BOOLEAN,
                            notes TEXT,
                            original_language TEXT,
                            original_title TEXT,
                            overview TEXT,
                            poster_path TEXT,
                            recent_change BOOLEAN,
                            release_date TEXT,
                            seasons_number INT,
                            soon_release BOOLEAN,
                            status TEXT,
                            tagline TEXT,
                            title TEXT,
                            watched BOOLEAN,
                            activate_notification BOOLEAN,
                            FOREIGN KEY (original_language) REFERENCES languages (iso_639_1)
                        );"""
            seasons_sql = """CREATE TABLE IF NOT EXISTS seasons (
                                episodes_number INTEGER,
                                id TEXT PRIMARY KEY,
                                number INTEGER,
                                overview TEXT,
                                poster_path TEXT,
                                title TEXT,
                                show_id INTERGER,
                                FOREIGN KEY (show_id) REFERENCES series (id) ON DELETE CASCADE
                            );"""
            episodes_sql = """CREATE TABLE IF NOT EXISTS episodes (
                                id TEXT PRIMARY KEY,
                                number INTEGER,
                                overview TEXT,
                                runtime INTEGER,
                                season_number INTEGER,
                                show_id INTEGER,
                                still_path TEXT,
                                title TEXT,
                                watched BOOLEAN,
                                FOREIGN KEY (show_id) REFERENCES series (id) ON DELETE CASCADE
                            );"""
            connection.cursor().execute(series_sql)
            connection.cursor().execute(seasons_sql)
            connection.cursor().execute(episodes_sql)
            connection.commit()

    @staticmethod
    def update_series_table() -> None:
        """
        Update the series table to add new data needed for notifications
        Args:
            None

        Returns:
            None
        """

        with sqlite3.connect(shared.db) as connection:

            sql = """pragma table_info(series)"""
            result = connection.cursor().execute(sql).fetchall()

            if not any(item[1] == "notes" for item in result):
                sql = """ALTER TABLE series
                            ADD notes TEXT
                            DEFAULT '';"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "last_air_date" for item in result):
                sql = """ALTER TABLE series
                            ADD last_air_date TEXT
                            DEFAULT '';"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "color" for item in result):
                sql = """ALTER TABLE series
                            ADD color BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "new_release" for item in result):
                sql = """ALTER TABLE series
                            ADD new_release BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "recent_change" for item in result):
                sql = """ALTER TABLE series
                            ADD recent_change BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "next_air_date" for item in result):
                sql = """ALTER TABLE series
                            ADD next_air_date TEXT
                            DEFAULT '';"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "soon_release" for item in result):
                sql = """ALTER TABLE series
                            ADD soon_release BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "activate_notification" for item in result):
                sql = """ALTER TABLE series
                            ADD activate_notification BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "last_episode_number" for item in result):
                sql = """ALTER TABLE series
                            ADD last_episode_number TEXT
                            DEFAULT ('');"""
                connection.cursor().execute(sql)
                connection.commit()

            connection.cursor().close()
            sql = """SELECT * FROM series;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql)
            connection.cursor().close()

            for entry in result:
                backdrop = entry["backdrop_path"]
                poster = entry["poster_path"]
                new_watched = (all(episode.watched for season in LocalProvider.get_all_seasons(
                    entry["id"]) for episode in season.episodes))

                if shared.DEBUG:  # if we are in debug build we add Devel to the path, this way we can copy release databases to the debug build to test.
                    backdrop = backdrop.replace(".Devel", '')
                    poster = poster.replace(".Devel", '')
                index = poster.find("/data")
                if index > 0:
                    if shared.DEBUG:
                        backdrop = backdrop[:index] + \
                            ".Devel" + backdrop[index:]
                        poster = poster[:index] + ".Devel" + poster[index:]

                    color = LocalProvider.compute_badge_color(Path(poster[7:]))
                else:
                    color = False

                activate_notification = entry["in_production"]

                sql = """UPDATE series SET activate_notification = ?, backdrop_path = ?, poster_path = ?, color = ?, watched = ? WHERE id = ?;"""
                result = connection.cursor().execute(sql, (
                    activate_notification,
                    backdrop,
                    poster,
                    color,
                    new_watched,
                    entry["id"],
                ))
                connection.commit()

    @staticmethod
    def update_movies_table() -> None:

        with sqlite3.connect(shared.db) as connection:

            sql = """pragma table_info(movies)"""
            result = connection.cursor().execute(sql).fetchall()

            if not any(item[1] == "notes" for item in result):
                sql = """ALTER TABLE movies
                            ADD notes TEXT
                            DEFAULT '';"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "color" for item in result):
                sql = """ALTER TABLE movies
                            ADD color BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "recent_change" for item in result):
                sql = """ALTER TABLE movies
                            ADD recent_change BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "activate_notification" for item in result):
                sql = """ALTER TABLE movies
                            ADD activate_notification BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "new_release" for item in result):
                sql = """ALTER TABLE movies
                            ADD new_release BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            if not any(item[1] == "soon_release" for item in result):
                sql = """ALTER TABLE movies
                            ADD soon_release BOOLEAN
                            DEFAULT (0);"""
                connection.cursor().execute(sql)
                connection.commit()

            sql = """SELECT * FROM movies;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql)
            connection.cursor().close()
            for entry in result:
                backdrop = entry["backdrop_path"]
                poster = entry["poster_path"]
                if shared.DEBUG:
                    backdrop = backdrop.replace(".Devel", '')
                    poster = poster.replace(".Devel", '')
                index = backdrop.find("/data")
                if index > 0:
                    if shared.DEBUG:
                        backdrop = backdrop[:index] + \
                            ".Devel" + backdrop[index:]
                        poster = poster[:index] + ".Devel" + poster[index:]
                    color = LocalProvider.compute_badge_color(Path(poster[7:]))
                else:
                    color = False

                if entry["release_date"]:
                    activate_notification = datetime.strptime(
                        entry["release_date"], '%Y-%m-%d') > datetime.now()
                else:
                    activate_notification = False

                sql = """UPDATE movies SET activate_notification = ?, backdrop_path = ?, poster_path = ?, color = ? WHERE id = ?;"""
                result = connection.cursor().execute(sql, (
                    activate_notification,
                    backdrop,
                    poster,
                    color,
                    entry["id"],))
                connection.commit()

    @staticmethod
    def compute_badge_color(poster_path: Path) -> bool:
        im = Image.open(poster_path)
        box = (im.size[0]-175, 0, im.size[0], 175)
        region = im.crop(box)
        median = ImageStat.Stat(region).median
        if sum(median) < 3 * 128:
            return True
        else:
            return False

    @staticmethod
    def create_languages_table() -> None:
        """
        Creates the table used to store the available languages in a local database.
        Not implemented yet.

        Args:
            None

        Returns:
            None
        """

        with sqlite3.connect(shared.db) as connection:
            logging.debug('[db] Create languages table')
            sql = """CREATE TABLE IF NOT EXISTS languages (
                        iso_639_1 TEXT PRIMARY KEY,
                        name TEXT NOT NULL
                     );"""
            connection.cursor().execute(sql)
            connection.commit()

    @staticmethod
    def create_tables() -> None:
        """
        Convenience method to create all tables with a single call.

        Args:
            None

        Returns:
            None
        """

        LocalProvider.create_movies_table()
        LocalProvider.create_series_table()
        LocalProvider.create_languages_table()

    @staticmethod
    def add_language(language: LanguageModel) -> int | None:
        """
        Inserts the provided LanguageModel in the languages table.

        Args:
            language: a LanguageModel to insert

        Returns:
            lastrowid: int or None containing the id of the last inserted row
        """

        with sqlite3.connect(shared.db) as connection:
            sql = 'INSERT INTO languages VALUES (?,?);'
            result = connection.cursor().execute(sql, (language.iso_name, language.name))
            connection.commit()
            logging.debug(f'[db] Add {language.name}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def add_movie(id: int = 0, movie: MovieModel | None = None) -> int | None:
        """
        Inserts a movie in the movies table, querying the data from TMDB if only id is provided.

        Args:
            id (int): tmdb id to query
            movie (MovieModel or None): movie to add

        Returns:
            int or None containing the id of the last inserted row
        """

        if not movie:
            movie = MovieModel(tmdb.get_movie(id))

        with sqlite3.connect(shared.db) as connection:
            sql = """INSERT INTO movies (
                activate_notification,
                add_date,
                backdrop_path,
                budget,
                color,
                genres,
                id,
                manual,
                new_release,
                original_language,
                original_title,
                overview,
                poster_path,
                recent_change,
                release_date,
                revenue,
                runtime,
                soon_release,
                status,
                tagline,
                title,
                watched,
                notes
                ) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"""
            result = connection.cursor().execute(sql, (
                movie.activate_notification,
                movie.add_date,
                movie.backdrop_path,
                movie.budget,
                movie.color,
                ','.join(movie.genres),
                movie.id,
                movie.manual,
                movie.new_release,
                movie.original_language.iso_name,  # type: ignore
                movie.original_title,
                movie.overview,
                movie.poster_path,
                movie.recent_change,
                movie.release_date,
                movie.revenue,
                movie.runtime,
                movie.soon_release,
                movie.status,
                movie.tagline,
                movie.title,
                movie.watched,
                movie.notes,
            ))
            connection.commit()
            logging.debug(
                f'[db] Add {movie.title}, {movie.release_date}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def add_series(id: int = 0, serie: SeriesModel | None = None) -> int | None:
        """
        Inserts a tv series in the series table, querying the data from TMDB if only id is provided.

        Args:
            id (int): tmdb id to query
            serie (SeriesModel or None): tv series to add

        Returns:
            int or None containing the id of the last inserted row
        """

        if not serie:
            serie = SeriesModel(tmdb.get_serie(id))

        with sqlite3.connect(shared.db) as connection:
            sql = """INSERT INTO series (
                activate_notification,
                add_date,
                backdrop_path,
                color,
                created_by,
                episodes_number,
                genres,
                id,
                in_production,
                last_air_date,
                last_episode_number,
                manual,
                next_air_date,
                new_release,
                original_language,
                original_title,
                overview,
                poster_path,
                recent_change,
                release_date,
                seasons_number,
                soon_release,
                status,
                tagline,
                title,
                watched,
                notes
            ) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"""
            result = connection.cursor().execute(sql, (
                serie.activate_notification,
                serie.add_date,
                serie.backdrop_path,
                serie.color,
                ','.join(serie.created_by),
                serie.episodes_number,
                ','.join(serie.genres),
                serie.id,
                serie.in_production,
                serie.last_air_date,
                serie.last_episode_number,
                serie.manual,
                serie.next_air_date,
                serie.new_release,
                serie.original_language.iso_name,  # type: ignore
                serie.original_title,
                serie.overview,
                serie.poster_path,
                serie.recent_change,
                serie.release_date,
                serie.seasons_number,
                serie.soon_release,
                serie.status,
                serie.tagline,
                serie.title,
                serie.watched,
                serie.notes,
            ))

            for season in serie.seasons:
                sql = 'INSERT INTO seasons VALUES (?,?,?,?,?,?,?);'
                connection.cursor().execute(sql, (
                    season.episodes_number,
                    season.id,
                    season.number,
                    season.overview,
                    season.poster_path,
                    season.title,
                    season.show_id
                ))

                for episode in season.episodes:
                    sql = 'INSERT INTO episodes VALUES (?,?,?,?,?,?,?,?,?);'
                    connection.cursor().execute(sql, (
                        episode.id,
                        episode.number,
                        episode.overview,
                        episode.runtime,
                        episode.season_number,
                        episode.show_id,
                        episode.still_path,
                        episode.title,
                        episode.watched
                    ))

            connection.commit()
            connection.cursor().close()
            logging.debug(
                f'[db] Add {serie.title}, {serie.release_date}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def add_content(id: int, media_type: str) -> int | None:
        """
        Convenience method to add movies and series from TDMB without using separate methods.

        Args:
            id (int): id of the content
            media_type (str): content's media type

        Returns:
            int or None containing the id of the last inserted row
        """

        if media_type == 'movie':
            return LocalProvider.add_movie(id)
        else:
            return LocalProvider.add_series(id)

    @staticmethod
    def get_language_by_code(iso_code: str) -> LanguageModel | None:
        """
        Retrieves a language from the db via its iso_639_1 code.

        Args:
            iso_code (str): iso_639_1 code of the language to look for

        Returns:
            LanguageModel of the requested language or None if not found in db
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM languages WHERE iso_639_1 = ?"""
            result = connection.cursor().execute(sql, (iso_code,)).fetchone()
            if result:
                language = LanguageModel(t=result)
                logging.debug(
                    f'[db] Get language by code {iso_code}: {language.name}')
                return language
            else:
                logging.error(f'[db] Get language by code {iso_code}: {None}')
                return None

    @staticmethod
    def get_movie_by_id(id: str) -> MovieModel | None:
        """
        Retrieves a movie from the db via its id.

        Args:
            id (str): id of the movie to look for

        Returns:
            MovieModel of the requested movie or None if not found in db
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM movies WHERE id = ?;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql, (id,)).fetchone()
            if result:
                movie = MovieModel(t=result)
                logging.debug(
                    f'[db] Get movie id {id}: {movie.title}, {movie.release_date}')
                return movie
            else:
                logging.error(f'[db] Get movie id {id}: None')
                return None

    @staticmethod
    def get_all_movies() -> List[MovieModel]:
        """
        Retrieves all movies from the db.

        Args:
            None

        Returns:
            List of MovieModel or None
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM movies;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql).fetchall()
            if result:
                logging.debug(f'[db] Get all movies: {result}')
                movies = []
                for movie in result:
                    movies.append(MovieModel(t=movie))
                return movies
            else:
                logging.debug(f'[db] Get all movies: {[]}')
                return []

    @staticmethod
    def mark_watched_movie(id: int, watched: bool) -> int | None:
        """
        Sets the watched flag on the movie with the provided id.

        Args:
            id (str): movie id to change
            watched (bool): status to set the flag to

        Returns:
            int or None containing the id of the last modified row
        """

        if watched:
            LocalProvider.set_new_release_status(id, False, movie=True)
            LocalProvider.set_soon_release_status(id, False, movie=True)

        with sqlite3.connect(shared.db) as connection:
            sql = """UPDATE movies SET watched = ? WHERE id = ?"""
            result = connection.cursor().execute(sql, (watched, id,))
            connection.commit()
            logging.debug(
                f'[db] Mark movie {id} watched {watched}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def delete_movie(id: str) -> int | None:
        """
        Deletes the movie with the provided id, removing associated files too.

        Args:
            id (str): movie id to delete

        Returns:
            int or None containing the id of the last modified row
        """

        logging.debug(f'[db] Movie {id}, delete requested')
        movie = LocalProvider.get_movie_by_id(id)

        if movie.backdrop_path.startswith('file'):  # type: ignore
            os.remove(movie.backdrop_path[7:])      # type: ignore
            logging.debug(f'[db] Movie {id}, deleted backdrop')

        if movie.poster_path.startswith('file'):    # type: ignore
            os.remove(movie.poster_path[7:])        # type: ignore
            logging.debug(f'[db] Movie {id}, deleted poster')

        with sqlite3.connect(shared.db) as connection:
            sql = """DELETE FROM movies WHERE id = ?"""
            result = connection.cursor().execute(sql, (id,))
            connection.commit()
            logging.debug(f'[db] Movie {id}, deleted: {result.lastrowid}')

        return result.lastrowid

    @staticmethod
    def get_all_seasons(show: str) -> List[SeasonModel]:
        """
        Retrieves metadata for all seasons of a show.

        Args:
            show (str): id of the show

        Returns:
            list of SeasonModel
        """

        seasons = []

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT *
                     FROM seasons
                     WHERE show_id = ?
                     ORDER BY number;"""

            results = connection.cursor().execute(sql, (show,)).fetchall()
            if results:
                logging.debug(f'[db] Get all seasons: {results}')
                for result in results:
                    seasons.append(SeasonModel(t=result))
            else:
                logging.debug(f'[db] Get all seasons: {[]}')

            return seasons

    @staticmethod
    def get_season_episodes(show: str, season_num: int) -> List[EpisodeModel]:
        """
        Retrieves the episodes for a season of a show.

        Args:
            show (str): id of the show
            season_num (int): season number

        Returns:
            list of EpisodeModel
        """

        episodes = []

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT *
                     FROM episodes
                     WHERE show_id = ? AND
                           season_number = ?
                     ORDER BY number;"""

            results = connection.cursor().execute(sql, (show, season_num,)).fetchall()
            if results:
                logging.debug(
                    f'[db] Get show {show} season {season_num}: {results}')
                for result in results:
                    episodes.append(EpisodeModel(t=result))
            else:
                logging.debug(
                    f'[db] Get show {show} season {season_num}: {[]}')

            return episodes

    @staticmethod
    def get_series_by_id(id: str) -> SeriesModel | None:
        """
        Retrieves the series with the provided id.

        Args:
            id (str): id of the series to retrieve

        Returns:
            SeriesModel for the requested series or None
        """

        with sqlite3.connect(shared.db) as connection:
            connection.row_factory = sqlite3.Row
            sql = 'SELECT * FROM series WHERE id=?;'
            result = connection.cursor().execute(sql, (id,)).fetchone()
            if result:
                serie = SeriesModel(t=result)
                logging.debug(
                    f'[db] Get tv serie id {id}: {serie.title}')
                return serie
            else:
                logging.error(f'[db] Get tv serie id {id}: None')
                return None

    @staticmethod
    def get_all_series() -> List[SeriesModel]:
        """
        Retrieves all tv series from the db.

        Args:
            None

        Returns:
            List of SeriesModel or None
        """
        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM series;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql).fetchall()
            if result:
                logging.debug(f'[db] Get all tv series: {result}')
                series = []
                for serie in result:
                    series.append(SeriesModel(t=serie))
                return series
            else:
                logging.debug(f'[db] Get all tv series: {[]}')
                return []

    @staticmethod
    def get_all_series_notification_list() -> List[SeriesModel]:
        """
        Retrieves all tv series from the notification_list.

        Args:
            None

        Returns:
            List of SeriesModel or None
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM series WHERE activate_notification = 1;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql).fetchall()
            if result:
                logging.debug(
                    f'[db] Get all tv series in notification_list: {result}')
                series = []
                for serie in result:
                    series.append(SeriesModel(t=serie))
                return series
            else:
                logging.debug(
                    f'[db] Get all tv series in notification_list: {[]}')
                return []

    @staticmethod
    def get_all_movies_notification_list() -> List[MovieModel]:
        """
        Retrieves all tv series from the notification_list.

        Args:
            None

        Returns:
            List of SeriesModel or None
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM movies WHERE activate_notification = 1;"""
            connection.row_factory = sqlite3.Row
            result = connection.cursor().execute(sql).fetchall()
            if result:
                logging.debug(
                    f'[db] Get all moviesin notification_list: {result}')
                movies = []
                for movie in result:
                    movies.append(MovieModel(t=movie))
                return movies
            else:
                logging.debug(
                    f'[db] Get all movies in notification_list: {[]}')
                return []

    @staticmethod
    def mark_watched_series(id: str, watched: bool) -> int | None:
        """
        Sets the watched flag on the specified series.

        Args:
            id (str): tv series id to change
            watched (bool): status to set the flag to

        Returns:
            int or None containing the id of the last modified row
        """

        # if all episodes are watched remove soon/new_release flags
        if watched:
            LocalProvider.set_new_release_status(id, False)
            LocalProvider.set_soon_release_status(id, False)

        with sqlite3.connect(shared.db) as connection:
            sql = 'UPDATE series SET watched = ? WHERE id = ?;'
            result = connection.cursor().execute(sql, (watched, id,))
            connection.commit()
            logging.debug(
                f'[db] Mark tv serie {id} watched {watched}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def delete_series(id: str) -> int | None:
        """
        Deletes the tv series with the provided id, removing associated files too.

        Args:
            id (str): tv series id to delete

        Returns:
            int or None containing the id of the last modified row
        """

        logging.debug(f'[db] TV series {id}, delete requested')
        series = LocalProvider.get_series_by_id(id)

        if series.backdrop_path and series.backdrop_path.startswith('file'):   # type: ignore
            try:
                os.remove(series.backdrop_path[7:])       # type: ignore
                logging.debug(f'[db] TV series {id}, deleted backdrop')
            except FileNotFoundError:
                pass

        if series.poster_path.startswith('file'):     # type: ignore
            try:
                os.remove(series.poster_path[7:])         # type: ignore
                logging.debug(f'[db] TV series {id}, deleted poster')
            except FileNotFoundError:
                pass

        if (shared.series_dir/id).is_dir():
            shutil.rmtree(shared.series_dir / id)
            logging.debug(
                f'[db] TV series {id}, deleted folder {shared.series_dir / id}')

        with sqlite3.connect(shared.db) as connection:
            connection.cursor().execute('PRAGMA foreign_keys = ON;')

            sql = """DELETE FROM series WHERE id = ?"""
            result = connection.cursor().execute(sql, (id,))
            connection.commit()
            logging.debug(f'[db] TV series {id}, deleted: {result.lastrowid}')

        return result.lastrowid

    @staticmethod
    def get_all_languages() -> List[LanguageModel]:
        """
        Retrieves all languages from the db.

        Args:
            None

        Returns:
            List of LanguageModel
        """
        with sqlite3.connect(shared.db) as connection:
            sql = 'SELECT * FROM languages ORDER BY iso_639_1'
            result = connection.cursor().execute(sql).fetchall()
            languages = []
            for language in result:
                languages.append(LanguageModel(t=language))
            logging.debug(f'[db] Get all languages: {result}')
            return languages

    @staticmethod
    def get_next_manual_movie() -> str:
        """
        Calculates the next id for a manually added movie.

        Args:
            None

        Returns:
            string with calculated id
        """

        with sqlite3.connect(shared.db) as connection:
            sql = "SELECT id FROM movies WHERE id LIKE 'M-%' ORDER BY id DESC;"
            result = connection.cursor().execute(sql).fetchone()
            if result:
                l = result[0].split('-')
                return f'{l[0]}-{int(l[1])+1}'
            else:
                return 'M-1'

    @staticmethod
    def get_next_manual_series() -> str:
        """
        Calculates the next id for a manually added tv series.

        Args:
            None

        Returns:
            string with calculated id
        """

        with sqlite3.connect(shared.db) as connection:
            sql = "SELECT id FROM series WHERE id LIKE 'M-%' ORDER BY id DESC;"
            result = connection.cursor().execute(sql).fetchone()
            if result:
                l = result[0].split('-')
                return f'{l[0]}-{int(l[1])+1}'
            else:
                return 'M-1'

    @staticmethod
    def get_next_manual_season() -> str:
        """
        Calculates the next id for a manually added season.

        Args:
            None

        Returns:
            string with calculated id
        """

        with sqlite3.connect(shared.db) as connection:
            sql = "SELECT id FROM seasons WHERE id LIKE 'M-%' ORDER BY id DESC;"
            result = connection.cursor().execute(sql).fetchone()
            if result:
                l = result[0].split('-')
                return f'{l[0]}-{int(l[1])+1}'
            else:
                return 'M-1'

    @staticmethod
    def get_next_manual_episode() -> str:
        """
        Calculates the next id for a manually added episode.

        Args:
            None

        Returns:
            string with calculated id
        """

        with sqlite3.connect(shared.db) as connection:
            sql = "SELECT id FROM episodes WHERE id LIKE 'M-%' ORDER BY id DESC;"
            result = connection.cursor().execute(sql).fetchone()
            if result:
                l = result[0].split('-')
                return f'{l[0]}-{int(l[1])+1}'
            else:
                return 'M-1'

    @staticmethod
    def get_language_by_name(name: str) -> LanguageModel | None:
        """
        Retrieves a language from the db via its name.

        Args:
            name (str):

        Returns:
            LanguageModel of the requested language or None if not found in db
        """

        with sqlite3.connect(shared.db) as connection:
            sql = 'SELECT * FROM languages WHERE name = ?;'
            result = connection.cursor().execute(sql, (name,)).fetchone()
            if result:
                language = LanguageModel(t=result)
                logging.debug(
                    f'[db] Get language by name {name}: {language.iso_name}')
                return language
            else:
                logging.error(f'[db] Get language by name {name}: {None}')
                return None

    @staticmethod
    def update_movie(old: MovieModel, new: MovieModel) -> int | None:
        """
        Updates a movie with new data.

        Args:
            old: movie to be updated
            new: new movie data

        Returns:
            int or None containing the id of the last modified row
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """UPDATE movies
                     SET 
                         backdrop_path = ?,
                         budget = ?,
                         color = ?,
                         genres = ?,
                         manual = ?,
                         original_language = ?,
                         original_title = ?,
                         overview = ?,
                         poster_path = ?,
                         release_date = ?,
                         revenue = ?,
                         runtime = ?,
                         status = ?,
                         tagline = ?,
                         title = ?,
                         notes = ?
                     WHERE id = ?;
                  """
            result = connection.cursor().execute(sql, (
                new.backdrop_path,
                new.budget,
                new.color,
                ','.join(new.genres),
                new.manual,
                new.original_language.iso_name,  # type: ignore
                new.original_title,
                new.overview,
                new.poster_path,
                new.release_date,
                new.revenue,
                new.runtime,
                new.status,
                new.tagline,
                new.title,
                old.notes,
                old.id,
            ))
            connection.commit()
            logging.debug(f'[db] Update movie {old.id}: {(new.backdrop_path, new.budget, ",".join(new.genres), new.manual, new.original_language.iso_name, new.original_title, new.overview, new.poster_path, new.release_date, new.revenue, new.runtime, new.status, new.tagline, new.title, old.notes, old.id)}')
        return result.lastrowid

    @staticmethod
    def update_series(old: SeriesModel, new: SeriesModel) -> int | None:
        """
        Updates a series with new data.

        Args:
            old: series to be updated
            new: new series data

        Returns:
            int or None containing the id of the last modified row
        """
        # Save episodes statuses before delete
        watched_episodes = []
        for season in old.seasons:  # type: ignore
            for episode in season.episodes:
                if episode.watched:
                    watched_episodes.append(episode.id)

        # remove series but not the posters, therefore not calling remove_series()
        # TODO Handle if the poster changes, the same problem in update_movie
        with sqlite3.connect(shared.db) as connection:
            connection.cursor().execute('PRAGMA foreign_keys = ON;')

            sql = """DELETE FROM series WHERE id = ?"""
            result = connection.cursor().execute(sql, (old.id,))
            connection.commit()
            connection.cursor().close()
            logging.debug(f'[db] TV series {id}, deleted: {result.lastrowid}')

        # Copy all flags that get reset but are still needed from the old to the new soon_release is automatically set in SeriesModel
        new.activate_notification = old.activate_notification
        new.add_date = old.add_date
        new.new_release = old.new_release
        new.recent_change = old.recent_change
        new.watched = old.watched
        new.notes = old.notes

        # Restore episodes statuses if they match before addition
        for idx, season in enumerate(new.seasons):
            for jdx, episode in enumerate(season.episodes):
                try:
                    watched_episodes.index(episode.id)
                    new.seasons[idx].episodes[jdx].watched = True
                except ValueError:
                    new.seasons[idx].episodes[jdx].watched = False
                    # we have found a new episode therefore the watched flag needs to be set to false
                    new.watched = False

        LocalProvider.add_series(serie=new)

        return result.lastrowid

    @staticmethod
    def mark_watched_episode(id: str, watched: bool) -> int | None:
        """
        Sets the watched flag on the specified episode.

        Args:
            id (str): episode id to change
            watched (bool): status to set the flag to

        Returns:
            int or None containing the id of the last modified row
        """
        with sqlite3.connect(shared.db, check_same_thread=False) as connection:
            sql = """UPDATE episodes SET watched = ? WHERE id = ?"""
            result = connection.cursor().execute(sql, (watched, id,))
            connection.commit()
            logging.debug(
                f'[db] Mark episode {id} watched {watched}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def get_episode_by_id(id: str) -> EpisodeModel | None:
        """
        Retrieves an episode from the db via its id.

        Args:
            id (str): id of the movie to look for

        Returns:
            EpisodeModel of the requested episode or None if not found in db
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """SELECT * FROM episodes WHERE id = ?;"""
            result = connection.cursor().execute(sql, (id,)).fetchone()
            if result:
                episode = EpisodeModel(t=result)
                logging.debug(f'[db] Get episode id {id}: {episode.title}')
                return episode
            else:
                logging.error(f'[db] Get episode id {id}: None')
                return None

    @staticmethod
    def set_notification_list_status(id: int, value: bool, movie: bool = False) -> None:
        """
        Sets notification_list status of the content with the id to value.

        Args:
            id (str): id of the movie to look for
            value (bool): the value to set

        Returns:
            None
        """
        if not movie:
            logging.debug(
                f'[db] TV series {id}, set activation_notification field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE series SET activate_notification = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,)).fetchone()
                connection.commit()
        else:
            logging.debug(
                f'[db] Movie {id}, set activation_notification field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE movies SET activate_notification = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,)).fetchone()
                connection.commit()

    @staticmethod
    def get_notification_list_status(id: int, movie: bool = False) -> bool:
        """
        Returns activate_notification status from the content with given id.

        Args:
            id (str): id of the content to look for

        Returns:
            Bool with the status
        """

        if not movie:
            logging.debug(f'[db] TV series {id}, get notification_list status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT activate_notification FROM series WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]
        else:
            logging.debug(f'[db] Movie {id}, get notification_list status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT activate_notification FROM movies WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]

    @staticmethod
    def set_new_release_status(id: int, value: bool, movie: bool = False) -> None:
        """
        Sets new_release status of the content with the id to value.

        Args:
            id (str): id of the content to look for
            value (bool): the value to set

        Returns:
            Success int or None if not found in db
        """
        if not movie:
            logging.debug(
                f'[db] TV series {id}, set new_release field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE series SET new_release = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,)).fetchone()
                connection.commit()
        else:
            logging.debug(f'[db] movie {id}, set new_release field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE movies SET new_release = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,)).fetchone()
                connection.commit()

    @staticmethod
    def get_new_release_status(id: int, movie: bool = False) -> None:
        """
        Returns new_release status from the content with given id.

        Args:
            id (str): id of the content to look for

        Returns:
            Success int or None if not found in db
        """

        if not movie:
            logging.debug(f'[db] TV series {id}, get new_release status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT new_release FROM series WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]
        else:
            logging.debug(f'[db] movie {id}, get new_release status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT new_release FROM movies WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]

    @staticmethod
    def set_soon_release_status(id: int, value: bool, movie: bool = False) -> None:
        """
        Sets soon_release status of the content with the id to value.

        Args:
            id (str): id of the content to look for
            value (bool): the value to set

        Returns:
            Success int or None if not found in db
        """

        if not movie:
            logging.debug(
                f'[db] TV series {id}, set soon_release field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE series SET soon_release = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,)).fetchone()
                connection.commit()
        else:
            logging.debug(
                f'[db] movie {id}, set soon_release field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE movies SET soon_release = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,)).fetchone()
                connection.commit()

    @staticmethod
    def get_soon_release_status(id: int, value: bool, movie: bool = False) -> None:
        """
        Returns soon_release status from the content with given id.

        Args:
            id (str): id of the content to look for

        Returns:
            Success int or None if not found in db
        """

        if not movie:
            logging.debug(f'[db] TV series {id}, get soon_release status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT soon_release FROM series WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]
        else:
            logging.debug(f'[db] movie {id}, get soon_release status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT soon_release FROM movies WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]

    @staticmethod
    def set_recent_change_status(id: int, value: bool, movie: bool = False) -> None:
        """
        Sets recent_change status of the content with the id to value.

        Args:
            id (str): id of the content to look for
            value (bool): the value to set

        Returns:
            Success int or None if not found in db
        """
        if not movie:
            logging.debug(
                f'[db] TV series {id}, set recent_change field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE series SET recent_change = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,))
                connection.commit()
        else:
            logging.debug(
                f'[db] movie {id}, set recent_change field to {value}')

            with sqlite3.connect(shared.db) as connection:
                sql = """UPDATE movies SET recent_change = ? WHERE id = ?"""
                result = connection.cursor().execute(sql, (value, id,))
                connection.commit()

    @staticmethod
    def get_recent_change_status(id: int, value: bool, movie: bool = False) -> None:
        """
        Returns recent_change status from the content with given id.

        Args:
            id (str): id of the content to look for

        Returns:
            Success int or None if not found in db
        """

        if not movie:
            logging.debug(f'[db] TV series {id}, get recent_change status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT recent_change FROM series WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]
        else:
            logging.debug(f'[db] movie {id}, get recent_change status')

            with sqlite3.connect(shared.db) as connection:
                sql = """SELECT recent_change FROM movies WHERE id = ?;"""
                result = connection.cursor().execute(sql, (id,)).fetchone()
                return result[0]

    @staticmethod
    def reset_recent_change() -> None:
        """
            Sets recent change of all content to False, must be called on application exit
        Args:
            None
        Returns:
            None
        """

        logging.debug(f'[db] All content, set recent_change to false')

        with sqlite3.connect(shared.db) as connection:
            sql = """UPDATE series SET recent_change = False;"""
            connection.cursor().execute(sql, ())
            connection.commit()
            sql = """UPDATE movies SET recent_change = False;"""
            connection.cursor().execute(sql, ())
            connection.commit()

    @staticmethod
    def reset_activate_notification() -> None:
        """
            Sets activate_notification of all content to False
        Args:
            None
        Returns:
            None
        """

        logging.debug(f'[db] All content, set activate_notification to false')

        with sqlite3.connect(shared.db) as connection:
            sql = """UPDATE series SET activate_notification = False;"""
            connection.cursor().execute(sql, ())
            connection.commit()
            sql = """UPDATE movies SET activate_notification = False;"""
            connection.cursor().execute(sql, ())
            connection.commit()

    @staticmethod
    def update_movie_notes(id: int, notes: str) -> int | None:
        """
        Updates the notes on the movie with the provided id.

        Args:
            id (int): movie id to change
            notes (str): notes to save

        Returns:
            int or None containing the id of the last modified row
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """UPDATE movies SET notes = ? WHERE id = ?"""
            result = connection.cursor().execute(sql, (notes, id,))
            connection.commit()
            logging.debug(
                f'[db] Save movie {id} notes {notes}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def update_serie_notes(id: int, notes: str) -> int | None:
        """
        Updates the notes on the TV Series with the provided id.

        Args:
            id (int): serie id to change
            notes (str): notes to save

        Returns:
            int or None containing the id of the last modified row
        """

        with sqlite3.connect(shared.db) as connection:
            sql = """UPDATE series SET notes = ? WHERE id = ?"""
            result = connection.cursor().execute(sql, (notes, id,))
            connection.commit()
            logging.debug(
                f'[db] Save TV Serie {id} notes {notes}: {result.lastrowid}')
        return result.lastrowid

    @staticmethod
    def export_data(export_path: Path) -> bool:
        """
        Exports all database content and associated images to a zip file.

        Args:
            export_path (Path): Path where to save the export zip file

        Returns:
            bool: True if export successful, False otherwise
        """
        try:
            with zipfile.ZipFile(export_path, 'w') as archive:
                # Export database content
                data: Dict[str, Any] = {
                    'movies': [],
                    'series': []
                }
                
                with sqlite3.connect(shared.db) as connection:
                    connection.row_factory = sqlite3.Row
                    
                    # Export movies
                    movies = connection.execute('SELECT * FROM movies').fetchall()
                    for movie in movies:
                        movie_dict = dict(movie)
                        if movie_dict['poster_path'].startswith('file://'):
                            poster_path = movie_dict['poster_path'][7:]
                            if Path(poster_path).exists():
                                relative_path = str(Path(poster_path)).split('data/')[-1]
                                archive.write(poster_path, f"images/{relative_path}")
                        if movie_dict['backdrop_path'].startswith('file://'):
                            backdrop_path = movie_dict['backdrop_path'][7:]
                            if Path(backdrop_path).exists():
                                relative_path = str(Path(backdrop_path)).split('data/')[-1]
                                archive.write(backdrop_path, f"images/{relative_path}")
                        data['movies'].append(movie_dict)

                    # Export series and related data
                    series = connection.execute('SELECT * FROM series').fetchall()
                    for serie in series:
                        serie_dict = dict(serie)
                        
                        # Add series poster/backdrop
                        if serie_dict['poster_path'].startswith('file://'):
                            poster_path = serie_dict['poster_path'][7:]
                            if Path(poster_path).exists():
                                relative_path = str(Path(poster_path)).split('data/')[-1]
                                archive.write(poster_path, f"images/{relative_path}")
                        if serie_dict['backdrop_path'].startswith('file://'):
                            backdrop_path = serie_dict['backdrop_path'][7:]
                            if Path(backdrop_path).exists():
                                relative_path = str(Path(backdrop_path)).split('data/')[-1]
                                archive.write(backdrop_path, f"images/{relative_path}")
                        
                        # Add seasons
                        seasons = connection.execute('SELECT * FROM seasons WHERE show_id = ?', 
                                                (serie_dict['id'],)).fetchall()
                        serie_dict['seasons'] = []
                        
                        for season in seasons:
                            season_dict = dict(season)
                            
                            # Add season poster with series/season folder structure
                            if season_dict['poster_path'] and season_dict['poster_path'].startswith('file://'):
                                poster_path = season_dict['poster_path'][7:]
                                if Path(poster_path).exists():
                                    filename = Path(poster_path).name
                                    archive.write(poster_path, 
                                            f"images/series/{serie_dict['id']}/{season_dict['number']}/{filename}")
                            
                            # Add episodes
                            episodes = connection.execute(
                                'SELECT * FROM episodes WHERE show_id = ? AND season_number = ?',
                                (serie_dict['id'], season_dict['number'])).fetchall()
                            season_dict['episodes'] = []
                            
                            for episode in episodes:
                                episode_dict = dict(episode)
                                
                                # Add episode still with series/season folder structure
                                if episode_dict['still_path'] and episode_dict['still_path'].startswith('file://'):
                                    still_path = episode_dict['still_path'][7:]
                                    if Path(still_path).exists():
                                        filename = Path(still_path).name
                                        archive.write(still_path,
                                                f"images/series/{serie_dict['id']}/{season_dict['number']}/{filename}")
                                
                                season_dict['episodes'].append(episode_dict)
                                
                            serie_dict['seasons'].append(season_dict)
                                
                        data['series'].append(serie_dict)

                # Write the JSON data
                archive.writestr('data.json', json.dumps(data, indent=2))
                
                return True
                
        except Exception as e:
            logging.error(f'[db] Export failed: {str(e)}')
            return False
        
    @staticmethod
    def import_data(import_path: Path) -> bool:
        """
        Imports database content and associated images from an export zip file.
        Restores images to their original locations and handles duplicate IDs.
        Overwrites existing entries if they have the same ID.

        Args:
            import_path (Path): Path to the zip file to import

        Returns:
            bool: True if import successful, False otherwise
        """
        try:
            logging.debug(f'[db] Starting import from {import_path}')
            
            with zipfile.ZipFile(import_path, 'r') as archive:
                # Read the JSON data
                json_data = archive.read('data.json').decode('utf-8')
                logging.debug(f'[db] Read JSON data: {json_data[:100]}...') # Log first 100 chars
                
                data = json.loads(json_data)
                if not data or 'movies' not in data or 'series' not in data:
                    logging.error(f'[db] Invalid JSON data in import file')
                    return False
                
                logging.debug(f'[db] Found {len(data["movies"])} movies and {len(data["series"])} series')
                
                with sqlite3.connect(shared.db) as connection:
                    
                    # Import movies
                    for movie in data['movies']:
                        # Handle movie images
                        if movie['poster_path'].startswith('file://'):
                            poster_filename = Path(movie['poster_path'][7:]).name
                            relative_path = str(Path(movie['poster_path'][7:])).split('data/')[-1]
                            target_dir = Path(shared.data_dir) / Path(relative_path).parent
                            target_dir.mkdir(parents=True, exist_ok=True)
                            
                            with archive.open(f"images/{relative_path}") as source, \
                                open(target_dir / poster_filename, 'wb') as target:
                                shutil.copyfileobj(source, target)
                                
                        if movie['backdrop_path'].startswith('file://'):
                            backdrop_filename = Path(movie['backdrop_path'][7:]).name
                            relative_path = str(Path(movie['backdrop_path'][7:])).split('data/')[-1]
                            target_dir = Path(shared.data_dir) / Path(relative_path).parent
                            target_dir.mkdir(parents=True, exist_ok=True)
                            
                            with archive.open(f"images/{relative_path}") as source, \
                                open(target_dir / backdrop_filename, 'wb') as target:
                                shutil.copyfileobj(source, target)

                        # Delete existing movie if present
                        cursor = connection.cursor()
                        sql = "SELECT COUNT(*) FROM movies WHERE id = ?"
                        if cursor.execute(sql, (movie['id'],)).fetchone()[0] > 0:
                            sql = "DELETE FROM movies WHERE id = ?"
                            connection.execute(sql, (movie['id'],))

                        # Handle manual movie IDs
                        if movie['id'].startswith('M-'):
                            movie['id'] = LocalProvider.get_next_manual_movie()

                        # Insert movie
                        columns = ', '.join(movie.keys())
                        placeholders = ', '.join(['?' for _ in movie])
                        sql = f"INSERT INTO movies ({columns}) VALUES ({placeholders})"
                        connection.execute(sql, list(movie.values()))

                    # Import series
                    for serie in data['series']:
                        seasons = serie.pop('seasons')
                        
                        # Handle series images
                        if serie['poster_path'].startswith('file://'):
                            poster_filename = Path(serie['poster_path'][7:]).name
                            relative_path = str(Path(serie['poster_path'][7:])).split('data/')[-1]
                            target_dir = Path(shared.data_dir) / Path(relative_path).parent
                            target_dir.mkdir(parents=True, exist_ok=True)
                            
                            with archive.open(f"images/{relative_path}") as source, \
                                open(target_dir / poster_filename, 'wb') as target:
                                shutil.copyfileobj(source, target)
                                
                        if serie['backdrop_path'].startswith('file://'):
                            backdrop_filename = Path(serie['backdrop_path'][7:]).name
                            relative_path = str(Path(serie['backdrop_path'][7:])).split('data/')[-1]
                            target_dir = Path(shared.data_dir) / Path(relative_path).parent
                            target_dir.mkdir(parents=True, exist_ok=True)
                            
                            with archive.open(f"images/{relative_path}") as source, \
                                open(target_dir / backdrop_filename, 'wb') as target:
                                shutil.copyfileobj(source, target)

                        # Delete existing series if present
                        connection.execute('PRAGMA foreign_keys = ON;')
                        cursor = connection.cursor()
                        
                        sql = "SELECT COUNT(*) FROM series WHERE id = ?"
                        if cursor.execute(sql, (serie['id'],)).fetchone()[0] > 0:
                            sql = "DELETE FROM episodes WHERE show_id = ?"
                            connection.execute(sql, (serie['id'],)) 
                            
                            sql = "DELETE FROM seasons WHERE show_id = ?"
                            connection.execute(sql, (serie['id'],))
                            
                            sql = "DELETE FROM series WHERE id = ?"
                            connection.execute(sql, (serie['id'],))

                        # Handle manual series IDs
                        if serie['id'].startswith('M-'):
                            serie['id'] = LocalProvider.get_next_manual_series()
                        series_id = serie['id']

                        # Insert series
                        columns = ', '.join(serie.keys())
                        placeholders = ', '.join(['?' for _ in serie])
                        sql = f"INSERT INTO series ({columns}) VALUES ({placeholders})"
                        connection.execute(sql, list(serie.values()))

                        # Import seasons and episodes
                        for season in seasons:
                            episodes = season.pop('episodes')
                            
                            # Handle season images
                            if season['poster_path'].startswith('file://'):
                                poster_filename = Path(season['poster_path'][7:]).name
                                target_dir = Path(shared.series_dir) / series_id / str(season['number'])
                                target_dir.mkdir(parents=True, exist_ok=True)
                                
                                image_path = f"images/series/{series_id}/{season['number']}/{poster_filename}"
                                with archive.open(image_path) as source, \
                                    open(target_dir / poster_filename, 'wb') as target:
                                    shutil.copyfileobj(source, target)

                            # Handle manual season IDs
                            if season['id'].startswith('M-'):
                                season['id'] = LocalProvider.get_next_manual_season()

                            # Insert season
                            columns = ', '.join(season.keys())
                            placeholders = ', '.join(['?' for _ in season])
                            sql = f"INSERT INTO seasons ({columns}) VALUES ({placeholders})"
                            connection.execute(sql, list(season.values()))

                            # Import episodes
                            for episode in episodes:
                                # Handle episode images
                                if episode['still_path'].startswith('file://'):
                                    still_filename = Path(episode['still_path'][7:]).name
                                    target_dir = Path(shared.series_dir) / series_id / str(season['number'])
                                    target_dir.mkdir(parents=True, exist_ok=True)
                                    
                                    image_path = f"images/series/{series_id}/{season['number']}/{still_filename}"
                                    with archive.open(image_path) as source, \
                                        open(target_dir / still_filename, 'wb') as target:
                                        shutil.copyfileobj(source, target)

                                # Handle manual episode IDs
                                if episode['id'].startswith('M-'):
                                    episode['id'] = LocalProvider.get_next_manual_episode()

                                # Insert episode
                                columns = ', '.join(episode.keys())
                                placeholders = ', '.join(['?' for _ in episode])
                                sql = f"INSERT INTO episodes ({columns}) VALUES ({placeholders})"
                                connection.execute(sql, list(episode.values()))

                    connection.commit()
                
            return True
                
        except Exception as e:
            logging.error(f'[db] Import failed: {str(e)}')
            return False